import bibtexparser
import os
import re
from typing import List, Dict, Any
import shutil


class BibTeXEntry:
    # To Debug:
    # keys = set()
    def __init__(self, bib_entry: Dict[str, Any]) -> None:
        self.entry_type: str = bib_entry['ENTRYTYPE']
        self.citation_key: str = bib_entry['ID']
        for key, value in bib_entry.items():
            if value is None or key in ['entry_type', 'citation_key', 'ENTRYTYPE', 'ID']:
                continue
            setattr(self, key, value)
            # To Debug:
            # BibTeXEntry.keys.add(key)

    def get_dict(self):
        res_dict = vars(self)
        if "entry_type" in res_dict:
            res_dict["ENTRYTYPE"] = res_dict.pop("entry_type")
        if "citation_key" in res_dict:
            res_dict["ID"] = res_dict.pop("citation_key")
        return res_dict

    def get(self, attr: str):
        try:
            if attr == "dict":
                return self.get_dict()
            return getattr(self, attr)
        except:
            return None

    def get_links(self):
        file = self.get("file")
        if not file:
            return None
        links = {}
        link_pattern = r'(?<!\\);'
        sub_pattern = r'([^:]+):((?:\\.|[^\\:^\\])*):([^:]+)'
        for group in re.split(link_pattern, file):
            for link in re.findall(sub_pattern, group):
                name, path, mime_type = link
                path = path.replace("\\:\\", ":\\")
                path = path.replace("\\\\", "\\")
                path = path.replace("\\;", ";")
                links.update({path.strip(): {
                    'name': name.strip(),
                    'mime_type': mime_type.strip()
                }
                })
        return links

    def set_links(self, links: Dict[str, str]):
        res = []
        for path, values in links.items():
            name = values["name"]
            mime_type = values["mime_type"]
            path = path.replace(";", "\\;")
            path = path.replace(":\\", "\\:\\\\")
            # path = path.replace("\\", "\\\\")
            res.append(f"{name}:{path}:{mime_type}")
        self.file = ";".join(res)

    def save(self, file_path: str = None, mode: str = 'w') -> str:
        res = "\n"
        res += f"@{self.entry_type}{{{self.citation_key},\n"
        for attribute, value in reversed(self.__dict__.items()):
            if value is None or attribute in ['entry_type', 'citation_key']:
                continue
            if isinstance(value, list):
                value = " and ".join(value)
            res += f"    {attribute} = {{{value}}},\n"
        res += "}\n"
        if file_path:
            with open(file_path, mode, encoding="utf-8") as file:
                file.write(res)
        return res


class BibResources:
    def __init__(self, path=None, entries: List[BibTeXEntry] = None) -> None:
        self.entries: Dict[str, BibTeXEntry] = entries or {}
        self.missing: Dict[str, BibTeXEntry] = {}
        self.folder_path = path or None
        self.files_folder_path = None
        self.bibtex_path = None
        self.pdf_path = None
        self.pdfs = []
        if path:
            self.from_path(path)

    def get_metadata(self, path: str):
        if path in self.pdfs:
            i = self.pdfs.index(path)
            basename = os.path.basename(os.path.dirname(path))
            if basename in self.entries:
                data = self.entries[basename].get_dict()
                return data
            else:
                print("Error adding metadata for " + path)

            # data = list(self.entries.values())[i].get_dict()

            # if data and 'file' in data:
            #     if path not in data['file']:
            #         print("Error adding metadata for " + path)
            # return data
        return None

    def get_pdf_path(self, path=None):
        if self.pdf_path and not path:
            return self.pdf_path
        if path is None:
            path = self.folder_path
        else:
            self.folder_path = path
        self.pdf_path = os.path.join(path, "00_PDFs")
        return self.pdf_path

    def add_entries(self):
        with open(self.bibtex_path, encoding="utf-8") as bibtex_file:
            parser = bibtexparser.bparser.BibTexParser()
            parser.ignore_nonstandard_types = False
            parser.homogenize_fields = False
            parser.common_strings = False
            bib_database = bibtexparser.load(bibtex_file, parser)
            for entry in bib_database.entries:
                value = BibTeXEntry(entry)
                key = value.citation_key
                self.entries[key] = value
        # To Debug:
        # print(BibTeXEntry.keys)

    def look_for_export(self, path, folder_name):
        for entry in os.listdir(path):
            if entry == folder_name and os.path.isdir(os.path.join(path, entry)):
                new_path = os.path.join(path, entry)
                self.look_for_export(new_path, folder_name)
                if self.files_folder_path and self.bibtex_path:
                    return
            # todo: make an implementation where the og bibtex file is written on
            elif entry == "files" and os.path.isdir(os.path.join(path, entry)):
                self.files_folder_path = os.path.join(path, entry)
            elif entry == folder_name + ".bib" and os.path.isfile(os.path.join(path, entry)):
                self.bibtex_path = os.path.join(path, entry)

    def from_path(self, path):
        self.get_pdf_path(path)
        folder_name = os.path.basename(path)
        self.look_for_export(path, folder_name)

        if not self.bibtex_path or not self.files_folder_path:
            print(f"Found no new Zotero export at {path}:")
            if not self.files_folder_path:
                print(f"There should be a folder called 'files'")
            if not self.bibtex_path:
                print(f"There must be a file called '{folder_name}.bib'")
                return None

        self.add_entries()
        self.sort_pdf()
        self.store_bibtex_entries(
            self.bibtex_path.replace(".bib", "_processed.bib"))

    def store_bibtex_entries(self, new_file_name=None) -> None:
        if new_file_name is None:
            new_file_name = self.bibtex_path
        with open(new_file_name, "w", encoding="utf-8") as file:
            for entry in self.entries.values():
                file.write(entry.save())

    def sort_pdf(self) -> None:
        for entry in self.entries.values():
            old_links = entry.get_links()
            if old_links is None:
                continue
            links = {}
            dst_folder = os.path.join(
                self.get_pdf_path(), entry.citation_key)

            failures: Dict[str, List] = {}
            for src_path, values in old_links.items():
                basename = os.path.basename(src_path)
                dst_path = os.path.join(dst_folder, basename)

                move = False
                failed = False
                if not os.path.isabs(src_path) and src_path.startswith("files"):
                    # relative -> copy
                    if self.files_folder_path:
                        prefix = self.files_folder_path
                        if src_path.startswith("files"):
                            prefix = os.path.dirname(prefix)
                        src_path = os.path.join(prefix, src_path)
                        move = True

                if os.path.exists(dst_path):
                    if dst_path not in links:
                        links[dst_path] = values
                        if dst_path.endswith(".pdf"):
                            self.pdfs.append(dst_path)
                    if move and os.path.exists(src_path):
                        os.remove(src_path)
                    continue

                if not os.path.exists(src_path):
                    failed = True

                if not failed:
                    os.makedirs(dst_folder, exist_ok=True)
                    if move:
                        try:
                            shutil.move(src_path, dst_path)
                        except Exception as e:
                            print(e)
                            failed = True
                    else:
                        try:
                            shutil.copy(src_path, dst_path)
                        except Exception as e:
                            print(e)
                            failed = True
                if not failed:
                    links[dst_path] = values
                    if basename in failures:
                        del failures[basename]
                    if dst_path.endswith(".pdf"):
                        self.pdfs.append(dst_path)
                else:
                    if basename not in failures:
                        failures[basename] = []
                    failures[basename].append(src_path)

            entry.set_links(links)
            for key, values in failures.items():
                print('Missing: ' + ' ; '.join(values))
                continue
        if self.files_folder_path:
            walk = list(os.walk(self.files_folder_path))
            for path, _, _ in walk[::-1]:
                if len(os.listdir(path)) == 0:
                    os.rmdir(path)
        walk = list(os.walk(self.get_pdf_path()))
        for path, _, _ in walk[::-1]:
            if len(os.listdir(path)) == 0:
                os.rmdir(path)
        print(f"We now have {len(self.pdfs)} PDFs stored at {self.pdf_path}")



import pdfminer
from pdfminer.high_level import extract_text, extract_pages
from pdfminer.layout import LAParams
from pdfminer.pdfinterp import PDFPageInterpreter, PDFResourceManager
from pdfminer.pdfpage import PDFPage
from pdfminer.converter import TextConverter
from pdfrw import PdfReader
from pdfquery import PDFQuery
# from tika import parser
import PyPDF2
import fitz
import io
import os
import re
import Levenshtein
from typing import List
import statistics
from collections import defaultdict
# does not (really) work on windows: from polyglot.detect import Detector


# https://stackoverflow.com/questions/21827874/timeout-a-function-windows
from threading import Thread
import functools


def timeout(timeout):
    def deco(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            res = [Exception('function [%s] timeout [%s seconds] exceeded!' % (
                func.__name__, timeout))]

            def newFunc():
                try:
                    res[0] = func(*args, **kwargs)
                except Exception as e:
                    res[0] = e
            t = Thread(target=newFunc)
            t.daemon = True
            try:
                t.start()
                t.join(timeout)
            except Exception as je:
                print('error starting thread')
                raise je
            ret = res[0]
            if isinstance(ret, BaseException):
                raise ret
            return ret
        return wrapper
    return deco


class PDFDocument:
    def __init__(self, path:str =None, text=None, pages=None, language=None, force=False):
        """
        A class representing a PDF document.

        Args:
            path (str): Path to the PDF document.
            text (str): Text content of the PDF document.
            pages (list): A list of strings, where each string is the text content of a single page.

        Attributes:
            path (str): Path to the PDF document.
            text (str): Text content of the PDF document.
            pages (list): A list of strings, where each string is the text content of a single page.
        """
        self.path:str = path
        self.text = text
        self.text_path = None
        self.pages = pages
        self.language = language
        self.metadata: dict = None
        if path:
            self.from_file(force=force)
            self.save(force=force)

    def get_manageable_text_sizes(self):
        # todo: make text container so this function has a better place to be stored just once
        text = self.text
        pieces = []
        jump = 50000

        if len(text) < jump * 2:
            pieces = [text]
        else:
            # Break text into smaller chunks so we have enough memory to solve it
            # FIXME: doesn't properly break
            candidates = ["\n\n", "\n", ".", "!", "?", ",", " "]

            begin = 0
            end = jump
            while end < len(text):
                next_stop = 0
                for fi in candidates:
                    pos = text[begin + jump:end + jump].find(fi)
                    if pos >= 0 and pos <= jump * 2:
                        next_stop = pos
                        break
                end = begin + jump + next_stop + 1
                pieces.append(text[begin:end])
                begin = end
            pieces.append(text[begin:])
        return pieces

    def add_metadata(self, data):
        if not self.metadata:
            self.metadata = data
        else:
            self.metadata.update(data)
        self.detect_language()

    def get_text_path(self):
        if self.text_path:
            return self.text_path
        if self.path:
            self.text_path = self.path.replace(".pdf", ".txt")
            return self.text_path
        else:
            return None

    def get_dict(self):
        return {
            'text': self.text,
        }

    def get(self, attr: str):
        if attr == "dict":
            return self.get_dict()
        elif attr == "title":
            return os.path.splitext(os.path.basename(self.path))[0]
        elif hasattr(self, attr):
            return getattr(self, attr)

    def save(self, force=False):
        if self.text:
            if force or not os.path.exists(self.text_path):
                with open(self.get_text_path(), "w", encoding="utf-8") as f:
                    f.write(self.text)

    def from_text_file(self):
        with open(self.text_path, "r", encoding="utf-8") as f:
            self.text = f.read()

    def has_text(self):
        if self.text and re.search('[a-zA-Z]', self.text):
            return True
        return False

    def from_file(self, path=None, force=False):
        """
        Load PDF document from the given path.

        Args:
            path (str): Path to the PDF document.
        """
        if not force and os.path.exists(self.get_text_path()):
            self.from_text_file()
            self.detect_language()
            if self.has_text():
                return
        if path is None:
            path = self.path
        else:
            self.path = path
        if not path:
            print("need path to load PDFDocument")
            return None
        # Ranking according to minor testing, future improvements welcome!

        error_path = self.get_text_path().replace(".txt", "_ERROR.txt")
        if not os.path.exists(error_path):
            try:
                if not self.has_text():
                    self.text = extract_text_pdfminer(path)
                if not self.has_text():
                    self.text = extract_text_pymupdf(path)
                if not self.has_text():
                    self.text = extract_text_pypdf2(path)
                if not self.has_text():
                    self.text = extract_text_pdfquery(path)
            except Exception as e:
                print(e)
        if not self.has_text():
            print("No way to extract text from: " + path)
            open(error_path, "a", encoding="utf8").close()

            # needs Java
            # self.text_tika = extract_text_tika(path)
            # looks like this can't do proper text
            # self.text_pdfrw = extract_text_pdfrw(path)
        self.clean_text()
        self.detect_language()

    def detect_language(self):
        proper = {
            "english": "en",
            "german": "de",
            "deutsch": "de",
        }
        # https://stackoverflow.com/questions/39142778/how-to-determine-the-language-of-a-piece-of-text
        # TODO: All these seem to involve further libraries like Visual C++ etc., and/or are tricky to run on Windows
        if not self.language:
            if self.metadata:
                keywords = ["langid", "language", "lid", "languageid"]
                for key in keywords:
                    if key in self.metadata:
                        value = self.metadata[key].lower()
                        if value in proper.values():
                            self.language = value
                            return
                        elif value in proper:
                            self.language = proper[value]
                            return
                        else:
                            print(f"{value} missing in language map")

            # todo: implement actual detection once issues are resolved
            else:
                self.language = "en"

    def clean_text(self, level=3):
        # https://tex.stackexchange.com/questions/33476/why-cant-fi-be-separated-when-being-copied-from-a-compiled-pdf
        changes = {
            "lvl1": {
                "ﬁ": "fi",
                "ﬀ": "ff",
                "ﬁ": "fi",
                "ﬂ": "fl",
                "ﬃ": "ffi",
                "ﬄ": "ffl",
            },
            "lvl2": {
                "–": "--",
                "—": "---"
            },
            "lvl3": {
                "“": "‘‘",
                "”": "’’",
                "¡": "!‘",
                "¿": "?‘",
            }
        }
        if not self.text:
            return None
        for swap in list(changes.values())[:level]:
            for wrong, right in swap.items():
                self.text = self.text.replace(wrong, right)
        # TODO: Remove lines that are not text, i.e. page numbers, Header, Footer, etc.
        # TODO: Remove line breaks mid-word
        # Optional: Remove Footnotes etc., though these may provide value

    def iscomplete(self):
        if self.path and self.text:
            return True
        return False

    def complete(self, path=None):
        if not self.iscomplete():
            self.from_file(path)
        return self.iscomplete()

    def get_dict(self):
        return {
            'text': self.text,
            'pages': self.pages,
        }

    def clean(self):
        if "\n" in self.text:
            self.text = self.text.replace("-\n", "")
            self.text = self.text.replace("\n", " ")


@timeout(30)
def extract_text_pypdf2(pdf_path):
    try:
        with open(pdf_path, 'rb') as pdf_file:
            pdf_reader = PyPDF2.PdfReader(pdf_file)
            text = ''
            for page_num in range(len(pdf_reader.pages)):
                page_obj = pdf_reader.pages[page_num]
                text += page_obj.extract_text()
            return text
    except Exception as e:
        print(e)
        return None


# @timeout(30)
# def extract_text_tika(pdf_path):
#     try:
#         parsed_pdf = parser.from_file(pdf_path)
#         return parsed_pdf['content']
#     except Exception as e:
#         print(e)
#         return None


@timeout(30)
def extract_text_pymupdf(pdf_path):
    try:
        text = ''
        with fitz.open(pdf_path) as pdf:
            for page in pdf:
                text += page.get_text()
        return text
    except Exception as e:
        print(e)
        return None


@timeout(30)
def extract_text_pdfrw(pdf_path):
    # looks like this can't export proper Text
    try:
        pdf = PdfReader(pdf_path)
        text = ''
        for page in pdf.pages:
            text += page.Contents.stream
        return text
    except Exception as e:
        print(e)
        return None


MIN_SUFFIX_LENGTH = 3
MAX_SUFFIX_LENGTH = 500
HISTORY_LENGTH = 5


def find_similar_ending(string_list: List[str], begin=MIN_SUFFIX_LENGTH, history=HISTORY_LENGTH, do_endings=True):
    numbers_regex = r"\d+"
    x = begin

    true_shared_threshold = len(string_list) / 3
    endings = defaultdict(None)
    banished = []
    while x < MAX_SUFFIX_LENGTH:
        possible_endings = defaultdict(list)
        if do_endings:
            candidates = [piece[-x:] for piece in string_list]
        else:
            candidates = [piece[:x] for piece in string_list]
        for idx, candidate in enumerate(candidates):
            if idx in banished:
                continue
            clean_candidate = re.sub(numbers_regex, "0", candidate)
            if clean_candidate:
                possible_endings[clean_candidate].append(idx)
        for ending, users in possible_endings.items():
            # FIXME: current iteration struggles to handle:
            # "beginnings", starting
            # "endings", ending
            # with different amounts of digits
            if len(users) < true_shared_threshold:
                for user in users:
                    if user in endings and endings[user] == ending:
                        pass
                    else:
                        endings[user] = x - 1 if x > begin else None
                        banished.append(user)
            else:
                for user in users:
                    endings[user] = ending
        x += 1
        if len(banished) == len(string_list):
            return endings


def remove_repetition(string_list: List[str], footer: bool = True, header: bool = True, merge: bool = True) -> str:
    result = ""
    endings = find_similar_ending(string_list, MIN_SUFFIX_LENGTH)
    openings = find_similar_ending(
        string_list, MIN_SUFFIX_LENGTH, do_endings=False)
    for i in range(len(string_list)):
        end_limit = -endings[i] if (i in endings and endings[i]) else None
        start_limit = openings[i] if i in openings else None
        result += string_list[i][start_limit:end_limit]
    return result
    # if header:
    # if merge:
    #     return


@timeout(30)
def extract_text_pdfminer(pdf_path, clean=False):
    # Rank 1: appears to be the best. Formatting is nice, accuracy works.
    # Tested: Can read multiple Columns correctly
    # Test pending: Scans
    if not clean:
        text = extract_text(pdf_path, laparams=LAParams())
        return text
    # todo: cleaning is messy right now.
    with open(pdf_path, 'rb') as fp:
        rsrcmgr = PDFResourceManager()
        retstr = io.StringIO()
        codec = 'utf-8'
        laparams = LAParams()
        device = TextConverter(
            rsrcmgr, retstr, codec=codec, laparams=laparams)
        interpreter = PDFPageInterpreter(rsrcmgr, device)

        pages = []
        for page in PDFPage.get_pages(fp):
            interpreter.process_page(page)
            pages.append(retstr.getvalue())
            retstr.truncate(0)
            retstr.seek(0)
        return remove_repetition(pages)


@timeout(30)
def extract_text_pdfquery(pdf_path):
    try:
        pdf = PDFQuery(pdf_path)
        pdf.load()
        text = ''
        for element in pdf.tree.iter():
            if element.text:
                text += element.text
        return text
    except Exception as e:
        print(e)
        return None


from ...core import helper
from publish.MediaWiki import mediaWiki

import re
import csv
import sys
import json
import os
import time
import numpy
import difflib
import shutil
import threading
import spacy
from pprint import pprint
from os import listdir
from os.path import isfile, join
from datetime import datetime
from os.path import exists
import os.path as path
from flair.data import Sentence
from flair.models import SequenceTagger

editfolder = 'edited'

def test(input_path, playlist_name, playlist_info, episodes_info, tagger, nlp):
    # make a sentence
    # sentence = Sentence('I love Berlin .')

    # run NER over sentence
    # tagger.predict(sentence)

    # print('The following NER tags are found:')

    # iterate over entities and print each
    # for entity in sentence.get_spans('ner'):
    #     print(entity)
    tokens = {}
    # TODO: Make sure input_path comes from the right place
    #
    # --- 1. Setup --- #
    # Log file
    # check what info is available

    # Read existing info
    playlist_info = {}
    episodes_info = {}

    # If major changes are made:

    edit_path = input_path + editfolder
    data_files, data_folders = helper.extract_file_folder(edit_path)

    jsons = []
    for filename in data_files:
        split_tup = os.path.splitext(edit_path + '\\' + filename)
        file_name = split_tup[0]
        file_extension = split_tup[1]
        if file_extension == '.json':
            jsons.append(filename)

    # Dictionary block:
    lexicon = {}
    lexicon_episodes = {}
    lexicon_ep_count = {}
    lexicon_ner = {}
    lexicon_ner_episode = {}
    lexicon_ner_full = {}

    with open(edit_path + '\\' + '_text.txt', "r", encoding="utf8") as f:
        lines = f.readlines()
    lexicon_mtime = None
    do_words = True
    # do_words = False
    if do_words:
        for idx, line in enumerate(lines):
            lexicon_episode = {}
            filename = jsons[idx]
            fileID = int(filename[:3])
            words = helper.wordify(line)
            for word in words:
                helper.nested_add(lexicon_episode, [word], 1)
                helper.nested_add(lexicon_ep_count, [word, fileID], 1)
            for key, value in lexicon_episode.items():
                helper.nested_add(lexicon, [key], value)
            helper.nested_add(lexicon_episodes, [fileID], sorted(
                lexicon_episode.items(), reverse=True, key=lambda item: item[1]))
            print(str(len(words)) + ' words in ' + filename, flush=True)
            # if idx % 50 == 0 or idx == len(lines)-1:
            if idx == len(lines)-1:
                lexicon = dict(sorted(lexicon.items()))
                helper.dict2json(lexicon, '_lexicon_alphabetical', edit_path)
                lexicon = {k: v for k, v in sorted(
                    lexicon.items(), reverse=True, key=lambda item: item[1])}
                helper.dict2json(lexicon,
                                 "_lexicon", edit_path)
                lexicon_ep_count = {k: v for k, v in sorted(
                    lexicon_ep_count.items(), reverse=True, key=lambda item: len(list(item[1].values())))}
                helper.dict2json(lexicon_ep_count,
                                 "_lexicon_ep_count", edit_path)

    else:
        lexicon_mtime = helper.json2dict(lexicon, "_lexicon", edit_path)

    if nlp:
        do_lemma = True
    else:
        do_lemma = False
    if do_lemma:
        lemmacon, lemma_dict = helper.lemmatize(lexicon, nlp)
        lemmacon = {k: v for k, v in sorted(
            lemmacon.items(), reverse=True, key=lambda item: sum(list(item[1].values())))}
        helper.dict2json(lemmacon,
                         "_lemmacon", edit_path)
        helper.dict2json(lemma_dict,
                         "_lemma_dict", edit_path)
    # Stemming
    # https://medium.com/@tusharsri/nlp-a-quick-guide-to-stemming-60f1ca5db49e
    # Cutting words down till the same for alle declinations remains
    # Lemmatization
    # Link
    # Wir können von allen wörter auf das gleiche kern-wort mappen
    # https://github.com/wartaal/HanTa
    # https://pypi.org/project/spacy/
    # https://stackoverflow.com/questions/57857240/ho-to-do-lemmatization-on-german-text

    do_sim = True
    # do_sim = False
    if do_sim:
        lexicon_similar = {}
        done = 0
        if lexicon_mtime:
            try:
                lexicon_similar_mtime = helper.json2dict(lexicon_similar, "_lexicon_similar", edit_path)
                if lexicon_similar_mtime > lexicon_mtime:
                    last = list(lexicon_similar.keys())[-1]
                    done = list(lexicon).index(last)
            except:
                pass
        upper_range = 2000
        max = int(len(lexicon))
        total = str(len(lexicon))
        lock = threading.Lock() 
        print_every_x = 100
        def compare(word, comp, lexicon_similar):
            # Cosine: seems to match almost anything
            # are_sim = helper.similar(word, comp, "cosine")
            # levenshtein: seems to be best fit.
            are_sim = helper.similar(word, comp, "levenshtein")
            # sequencematcher: 
            # are_sim = helper.similar(word, comp, "sequencematcher")
            # jaccard: 
            # are_sim = helper.similar(word, comp, "jaccard")
            if are_sim:
                # print(word + " ~= " + comp, flush=True)
                with lock:  # acquire the lock
                    if word in lexicon_similar:
                        lexicon_similar[word] += [comp]
                    else:
                        lexicon_similar[word] = [comp]
        words = list(lexicon.keys())[:max]
        start_time = time.time()
        lap_time = start_time
        for idx, word in enumerate(words):
            if idx < done:
                continue
            # They we'd neet to lock
            threads = []
            for comp in words[idx+1:]:
                compare(word, comp, lexicon_similar)
            #     t = threading.Thread(target=compare, args=(word, comp, lexicon_similar))
            #     t.start()  # start the thread
            #     threads.append(t)  # add the thread to the list
            # for t in threads:
            #     t.join()  # wait for all threads to finish
            # if upper_range + 1 < max:
            #     upper_range+=1
            if idx % print_every_x == 0:
                # TODO: Include how long this took (per word)
                time_per_word = round((time.time() - lap_time) / print_every_x, 2) 
                lap_time = time.time()
                print(f"{str(idx)}/{total} words checked with currently {time_per_word} seconds per word.", flush=True)
                helper.dict2json(
                    lexicon_similar, "_lexicon_similar", edit_path)

    if tagger:
        do_ner = True
    else:
        do_ner = False
    if do_ner:
        for idx, line in enumerate(lines):
            lexicon_ner_episode = {}
            filename = jsons[idx]
            fileID = int(filename[:3])
            print(filename + ':')
            sentences = helper.sentencify(line)
            for s in sentences:
                sentence = Sentence(s)

                # run NER over sentence
                tagger.predict(sentence)

                # iterate over entities and print each
                sentence.get_token
                for entity in sentence.get_spans('ner'):
                    text = entity.text
                    helper.nested_add(lexicon_ner_episode, [text, fileID], 1)
                    helper.nested_add(lexicon_ner_full, [
                        text, fileID, entity.tag], [entity.score])
            for text, dict_var in lexicon_ner_episode.items():
                for fileID, value in dict_var.items():
                    helper.nested_add(lexicon_ner, [text, fileID], value)
            if idx % 50 == 0 or idx == len(lines)-1:
                lexicon_ner = dict(sorted(lexicon_ner.items()))
                lexicon_ner_full = dict(sorted(lexicon_ner_full.items()))
                helper.dict2json(lexicon_ner, '_lexicon_ner', edit_path)
                helper.dict2json(lexicon_ner_full,
                                 '_lexicon_ner_full', edit_path)

            print('', flush=True)


def main(input_path=os.getcwd(), playlist_name='BMZ', playlist_info={}, episodes_info={}):
    if input_path[-1] != '\\':
        input_path += '\\'
    # print('vorher', flush=True)
    # # log_file = open("logfile_" + playlist_name + ".log", "w", encoding='utf8')
    # # sys.stdout = log_file
    # print('nachher', flush=True)
    # load the NER tagger
    tagger = False
    nlp = False
    # tagger = SequenceTagger.load('ner-multi-fast')
    nlp = spacy.load('de_core_news_md')
    test(input_path, playlist_name, playlist_info, episodes_info, tagger, nlp)


if __name__ == '__main__':
    main()



import re

hashtag = '#'

def write_individual_wiki_page_BMZ(ID, episode_info):
    date = ''
    if 'date' in episode_info:
        date = episode_info['date']
    runtime = ''
    if 'runtime' in episode_info:
        runtime = episode_info['runtime']
    url = ''
    if 'url' in episode_info:
        url = episode_info['url']
    title_full = ''
    if 'title_full' in episode_info:
        title_full = episode_info['title_full']
    title = ''
    if 'title' in episode_info:
        title = episode_info['title']
    category = ''
    if len(title_full) > len (title):
        category = title_full[:-(len(title)+1)]
    transcript = ''
    if 'transcript' in episode_info:
        transcript = episode_info['transcript']
    leading_zeros = ''
    if ID < 9:
        leading_zeros = '00'
    elif ID < 99:
        leading_zeros = '0'

    description = ""
    if 'description' in episode_info:
        description += episode_info['description'] + '\n\n'
    text = f"""{description}"""
    #todo : continue putting all of this into the """string
    text += '{| class="wikitable"' + '\n'
    text += '! ' + hashtag + ' !! date !! runtime !! url' + '\n'
    text += '|-' + '\n'
    text += '| ' + str(ID+1) + ' || ' + date + ' || ' + \
        runtime + ' || ' + url + '\n'
    text += '|}' + '\n'
    text += '\n'
    text += '= Zusammenfassung =' + '\n'
    text += '<Zusammenfassung>' + '\n'
    text += '\n'
    text += '= Lessons Learned =' + '\n'
    text += '== <Lesson Learned 1> ==' + '\n'
    text += '\n'
    text += '= Transkript =' + '\n'
    text += transcript + '\n'
    text += '\n'
    # nickname = row[elements['title']].split("|", 1)[1]
    # nickname = nickname.replace("]]", "")
    # text += '[[Kategorie:Hagrids Hütte:Hörspiel|' + nickname + ']]' + '\n'
    text += '[[Kategorie:' +  category + '|' + leading_zeros + title + ']]'# + '\n'
    return text
    # filename = helper.fill_digits(id, 4)
    # with open(input_path + filename + '_wiki.txt', 'w') as f:
    #     f.write(text)

def write_mainpage(episodes_info):
    text = 'https://www.youtube.com/c/OnkelBarlow\n'
    text += '\n'
    text += '= Podcast =\n'
    text += '*[[OnkelBarlow/BMZ]]\n'
    text += '**[[:Category:OnkelBarlow/BMZ|BMZ Kategorie]]\n'
    text += '*[[:Category:Onkel Barlow:Guide|Guide]]\n'
    text += '\n'
    text += '= Übersicht =\n'
    text += '== Podcast ==\n'
    text += '{| class="wikitable sortable"\n'
    text += '|+ playlist\n'
    text += '|-\n'
    text += '! title !! description !! author_type !! author_name !! publisher !! language !! @type !! accessMode !! url !! image\n'
    text += '|-\n'
    text += '| [[OnkelBarlow/BMZ|BMZ]] || Listen to BMZ on Spotify.  || Person || [https://www.youtube.com/@OnkelBarlow OnkelBarlow] || George Zaal || de || PodcastSeries || auditory || [https://www.youtube.com/playlist?list=PLv3Aqr0YYT6S_8mdaKCBu0pskPLNEPSlJ YouTube][https://open.spotify.com/show/55UqBIZCSXoBZB5jVXEzHq Spotify] || [https://i.scdn.co/image/13d386beb25f08329057330ebe053172e6f7a506 image]\n'
    text += '|}\n'
    text += '\n'
    text += '\n'
    text += '\n'
    text += '=== Episoden ===\n'
    text += '{| class="wikitable sortable"\n'
    text += '|+ episodes\n'
    text += '|-\n'
    text += f'! {hashtag} !! date !! runtime !! title !! url\n'
    text += '|-\n'
    
    for ID, episode_info in episodes_info.items():
        date = ''
        if 'date' in episode_info:
            date = episode_info['date']
        runtime = ''
        if 'runtime' in episode_info:
            runtime = episode_info['runtime']
        url = ''
        if 'url' in episode_info:
            url = episode_info['url']
        link_to_title = ''
        if 'link_to_title' in episode_info:
            link_to_title = episode_info['link_to_title']

        text += '|-' + '\n'
        text += '| ' + str(ID+1) + ' || ' + date + ' || ' + \
            runtime + ' || ' + link_to_title +  ' || ' + url + '\n'


    text += '|}\n'
    text += '\n'
    text += '= Tags =\n'
    text += '[[Category:Podcast]]\n'
    text += '[[Category:German]]\n'
    text += '[[Category:Unterhaltung]]\n'
    text += '[[Category:Bildung]]\n'
    return text



from ...core import helper
from ...publish.Obsidian import obsidize
from ...publish.MediaWiki import mediaWiki
from ...extract.nlp import nlp

import re
import csv
import sys
import json
import os
import time
import numpy
from pprint import PrettyPrinter
from os import listdir
from os.path import isfile, join
from os.path import exists
import os.path as path
# Define
tokenfolder = helper.TOKENFOLDER
editfolder = helper.EDITFOLDER
# If major changes have been made: Flush out these old ones

# ---- Functions ---- #


# Convert extracted time to readable time


# Sort a dict according to a list.
# reminder: items not in the list won't be transferred.


# convert from Spotify index to dictionary


# convert from json to csv


def json2csv(playlist_info, episodes, input_path, playlist_name):
    # ---- Write ---- #
    playlist_header = [] + list(playlist_info.keys())
    episode_header = ["#"] + list(episodes[0].keys())
    # TODO: make sure to catch all, if first hasnt got all
    # output_path = input_path.replace('input','output')
    # You will need 'wb' mode in Python 2.x
    with open(input_path + 'playlist.csv', 'w', newline='', encoding='ANSI') as f:
        writer = csv.writer(f, delimiter=';')
        writer.writerow(playlist_header)
        values = []
        for key, value in playlist_info.items():
            values.append(value)
        writer.writerow(values)

    # You will need 'wb' mode in Python 2.x
    with open(input_path + 'episodes.csv', 'w', newline='', encoding='ANSI') as f:
        writer = csv.writer(f, delimiter=';')
        writer.writerow(episode_header)
        for key, value in episodes.items():
            values = [key+1]
            for ekey, evalue in value.items():
                actual_key = len(values)
                while ekey != episode_header[actual_key]:
                    values.append('')
                    actual_key += 1
                values.append(evalue)
            writer.writerow(values)
    return


def rreplace(s, old, new, occurrence):
    li = s.rsplit(old, occurrence)
    return new.join(li)


def clean_filenames():
    #! not foolproof yet
    return
    # for filename in data_files_audio:
    #     clean_name = re.sub('-[-\d\w]{11}.', '.', filename)
    #     path_old = input_path + '\\' + foldername + '\\' + filename
    #     path_new = input_path + '\\' + foldername + '\\' + clean_name
    #     if filename != clean_name:
    #         file_exists = exists(path_new)
    #         counter = 1
    #         while file_exists:
    #             split_tup = os.path.splitext(path_new)
    #             file_name = split_tup[0]
    #             file_extension = split_tup[1]
    #             if counter >1:
    #                 file_name = rreplace(file_name,' #' + str(counter), '', 1)
    #             counter += 1
    #             path_new = file_name + " #" + str(counter) + file_extension
    #             file_exists = exists(path_new)
    #         os.rename(path_old, path_new)


token_puncs = {
    '......':   r'(......)',
    '...':   r'(...$)',
    ',':   r'(,$)',
    '.':   r'((?<!\d)\.$)',
    '!':   r'(!$)',
    '?':   r'(\?$)',
}

token_puncs_ok = {
    '.':   r'\d\.$',
}


def split_punctuation(token):
    found = []
    for find, finder in token_puncs.items():
        if find in token:
            token_parts = re.split(finder, token, 1)
            for part in token_parts:
                if part == '':
                    continue
                if part == find:
                    found.append(part)
                if part == token:
                    continue
                else:
                    found += split_punctuation(part)
    if not found:
        found = [token]
    return found


def dictify_tokens(token, token_text, var_int, var_dict):
    if token in var_dict:
        if token_text in var_dict[token]:
            if var_int in var_dict[token][token_text]:
                var_dict[token][token_text][var_int] += 1
            else:
                var_dict[token][token_text].update({var_int: 1})
        else:
            var_dict[token][token_text] = {var_int: 1}
    else:
        var_dict[token] = {token_text: {var_int: 1}}


def do_token_stuff(input_path, jsons):
    # phase 1: Addapt names
    detected_nothing = "detected_nothing!!!"
    detected_occurence = 'detected_occurence!!!'
    tokenpath = input_path + tokenfolder + '\\'

    token2string_dict = {}

    # Read in all the transcribts
    phase_2 = True
    if phase_2:
        # Todo: Build a Dict with { Token : Word or ,.!? }
        # Status: Currently on hold due to it Tokens not alligning 100% with words
        full_dict = {}
        wrong_dict = {}
        list_wrong_matches = []
        counter_match = 0
        counter_no_match = 0
        for filename in jsons:
            transcript = helper.get_transcript(input_path, filename)
            for segment in transcript['segments']:
                tokens = segment['tokens']
                text = segment['text']
                tokens_text = text.split()
                if len(tokens) != len(tokens_text):
                    better_tokens_text = []
                    for token in tokens_text:
                        temp = split_punctuation(token)
                        for x in temp:
                            better_tokens_text.append(x)
                    tokens_text = better_tokens_text
                if len(tokens) != len(tokens_text):
                    counter_no_match += 1
                    list_wrong_matches.append([tokens, tokens_text])
                else:
                    token_translation = {}
                    for idx, token in enumerate(tokens):
                        # dictify_tokens(token, tokens_text[idx], int(filename.split('_')[0]), token_translation)
                        dictify_tokens(token, tokens_text[idx], int(
                            filename.split('_')[0]), full_dict)
                    counter_match += 1
                # ['Hallo', 'und', 'herzlich', 'willkommen', 'hier', 'ist', 'Bardo', 'mit', 'einem', 'neuen', 'Video', 'zu', 'WoW', 'zu', ...]
                # ' Hallo und herzlich willkommen
                # [21242, 674, 45919, 46439,
                # hier ist Bardo mit
                # 3296, 1418, 363, 12850,
                # einem neuen Video zu
                # 2194, 6827, 21387, 9777,
                # WoW zu Legion heute'
                # 2164, 6622, 54 2164
                # ?? ??
                # 33024 9801
            if not os.path.exists(tokenpath):
                os.makedirs(tokenpath)
            # helper.dict2json(token_translation, filename.replace('.json', '') + "_tokens", tokenpath)

            print('Match rate: ' +
                  str(round(counter_match/counter_no_match, 2)), flush=True)
            # print('While ' + str(counter_match) + ' could be matched, ' + str(counter_no_match) + ' could not.', flush=True)
        for token in full_dict:
            if len(full_dict[token]) == 1:
                token2string_dict.update({token: list(full_dict[token])[0]})
        token2string_dict = dict(sorted(token2string_dict.items()))
        helper.dict2json(token2string_dict, "_token2string", tokenpath)
        helper.dict2json(full_dict, "_tokens", tokenpath)
        occurs_with = {}

        for tokens, tokens_text in list_wrong_matches:
            for token in tokens:
                if token in full_dict:
                    text = list(full_dict[token])[0]
                    if text in tokens_text:
                        tokens_text.remove(text)
                        tokens.remove(token)
            # print(',\t'.join(str(e) for e in tokens))
            # print(',\t'.join(str(e) for e in tokens_text), flush=True)
            while len(tokens) > len(tokens_text):
                tokens_text.append(detected_nothing)
            while len(tokens) < len(tokens_text):
                tokens.append(-1)
            for idx, token in enumerate(tokens):
                dictify_tokens(token, tokens_text[idx], int(
                    filename.split('_')[0]), wrong_dict)
                dictify_tokens(token, token, detected_occurence, occurs_with)
                for text in tokens_text:
                    dictify_tokens(token, token, text, occurs_with)
    else:
        with open(tokenpath + '_tokens_occurs_with.json', encoding='utf-8') as json_file:
            occurs_with = json.load(json_file)

    phase_3 = True
    if phase_3:
        # helper.dict2json(occurs_with, "_tokens_occurs_with", tokenpath)
        determined_tokens = {}
        progress_made = True
        while progress_made:
            progress_made = False
            blacklist = []
            occurs_with_filtered = {}
            # TODO: remove soon:
            for token in list(occurs_with):
                if detected_occurence in occurs_with[token][token]:
                    del occurs_with[token][token][detected_occurence]
                if detected_nothing in occurs_with[token][token]:
                    del occurs_with[token][token][detected_nothing]
                # make that work later on
                # if detected_occurence in occurs_with[token][token]:
                #     det_count = occurs_with[token][token][detected_occurence]
                # else:
                max_v = 0
                for key, value in occurs_with[token][token].items():
                    max_v = max(max_v, value)
                det_count = max_v
                occurs_with_filtered[token] = {}

                for text in occurs_with[token][token]:
                    fin_count = occurs_with[token][token][text]
                    if fin_count >= 0.5 * det_count:
                        occurs_with_filtered[token][text] = fin_count
                if len(occurs_with_filtered[token]) == 1:
                    text = list(occurs_with_filtered[token])[0]
                    determined_tokens[token] = text
                    blacklist.append([token, text])
                    progress_made = True
            for token, text in blacklist:
                if token in occurs_with:
                    del occurs_with[token]
                for token in list(occurs_with):
                    if text in occurs_with[token][token]:
                        del occurs_with[token][token][text]
            blacklist = []

        helper.dict2json(occurs_with_filtered,
                         "_tokens_occurs_with_filtered", tokenpath)
        helper.dict2json(determined_tokens, "_tokens_determined", tokenpath)

    phase_4 = True
    if phase_4:
        pass
    # helper.dict2json(occurs_with_filtered, "_tokens_occurs_with_filtered", tokenpath)
    # helper.dict2json(occurs_with, "_tokens_occurs_with", tokenpath)
    # helper.dict2json(wrong_dict, "_tokens_wrong", tokenpath)


def convert_to_wiki(input_path, playlist_name, playlist_info, episodes_info):
    playlist_info, episodes_info = helper.setup_infos(
        playlist_info, episodes_info, input_path)
    return mediaWiki.main(playlist_info, episodes_info, input_path, playlist_name)


def main():
    # path of our playlist folders: data_path -> str
    # names of our playlists: playlist_names -> List
    data_path, playlist_names = helper.get_data_folders()
    for pl_n in playlist_names:
        # Change stdout so it writes to logfile_{pl_n}.log
        old_stdout = sys.stdout
        log_file = open("logfile_" + pl_n + ".log", "w", encoding='utf8')
        sys.stdout = log_file

        # Setup
        data_pl_path = data_path + pl_n + '\\'
        playlist_info = {}
        episodes_info = {}

        # todo: read json file first
        wiki_episodes_info = convert_to_wiki(
            data_pl_path, pl_n, playlist_info, episodes_info)

        sys.stdout = old_stdout
        log_file.close()


if __name__ == '__main__':
    main()



MY_STOP_WORDS = {
    "en": set(
        """
like yeah know thing oh ok
""".split()
    ),
    "de": set(
        """
oh ok
mal einfach sagen eigentlich finden genau
quasi halt hey irgendwie bisschen bissch
beispielsweise irgendjemand all worum
""".split()
    )
}



import yaml
import os
import sys
import importlib


bnw_tools_spec = importlib.util.find_spec('bnw_tools')

if bnw_tools_spec is None:
    sys.path.append(os.getcwd())
    import bnw_tools

from bnw_tools.publish import util_wordcloud
from bnw_tools.publish.Obsidian import nlped_whispered_folder
from bnw_tools.extract.nlp import util_nlp
from bnw_tools.extract import util_whisper

# test = "C:/Users/TimWittenborg/workspace/test/whispered"
# test = "C:/Users/TimWittenborg/workspace/data"

query = [
    # {"folder": "D:/workspace/raw (MP3)/Weltverbesserer",
    #  "image": "D:/workspace/raw (MP3)/Weltverbesserer/mask/49faa272-1663-44cb-ae49-6c7a7356cc12 - Kopie - Kopie.jpg",
    #  "language": "de"},
    {"folder": "D:/workspace/raw (MP3)/HoP",
     "image": "D:/workspace/raw (MP3)/HoP/mask/ab67656300005f1fcabcff3dcfa8fbd5f5b0fa04.jpeg",
     "language": "en"},
]


# Extract
## Transcribe (Whisper)
nlptools = util_nlp.NLPTools()
for do in query:

    util_whisper.extract_info(do["folder"])

    ## NLP (Flair and SpaCy)
    folder = util_nlp.Folder(do["folder"], nlptools=nlptools, language=do["language"])

    # Publish

    # Wordcloud
    util_wordcloud.generate_wordcloud(
        folder.bag_of_words.get(), os.path.join(do["folder"], "00_bag_of_words"))
    util_wordcloud.generate_wordcloud(
        folder.named_entities.get_frequencies(), os.path.join(do["folder"], "00_named_entities"))
    mask = util_wordcloud.generate_mask(do["image"])
    util_wordcloud.generate_wordcloud_mask(
        folder.bag_of_words.get(), mask, os.path.join(do["folder"], "00_bag_of_words_mask"))
    util_wordcloud.generate_wordcloud_mask(folder.named_entities.get_frequencies(
    ), mask, os.path.join(do["folder"], "00_named_entities_mask"))

    # Obsidian
    nlped_whispered_folder.folder(do["folder"], force=True)



"""
Goal:
Calculate the "k_score" for a given set of keywords

Setup:
1: Run and document your search query.
2: Store your results in Zotero (or similar)
3: Export your results (incl. files) to Obsidian (or anywhere, really)
4: Setup the config
"""
import yaml
import os
import sys

# todo: make this import cleaner
sys.path.append(os.getcwd())
from publish.Obsidian import bibtex_to_notes

mypath = os.getcwd()
runpath = os.path.join(os.path.dirname(__file__), "runs")
runs = [f for f in os.listdir(
    runpath) if f.endswith(".yml")]
for run in runs:
    with open(os.path.join(runpath, run), "r") as ymlfile:
        cfg = yaml.safe_load(ymlfile)
    bibtex_to_notes.main(cfg)



import json
import difflib
import difflib
import Levenshtein
import re
import re
import os
from os import listdir
from os.path import isfile, join
from datetime import datetime
import math
from collections import Counter
import numpy as np
from numpy.linalg import norm
import csv

SIMILAR_ENOUGH = {
    "cosine": 0.9,
    "levenshtein": 0.85,
    "jaccard": 0.9,
    "sequencematcher": 0.9,
}
AUDIOFOLDER = 'mp3'
TOKENFOLDER = 'token'
EDITFOLDER = 'edited'


# Prepare dict so it can be used for MediaWiki
noFileChars = '":\<>*?/'


def write_to_py(code, path):
    with open(path, "w", newline="", encoding='UTF-8') as f:
        f.write(code)


def write_to_csv(column_names, rows, path, delimiter=","):
    with open(path, "w", newline="", encoding='UTF-8') as f:
        if type(rows[0]) == dict:
            writer = csv.DictWriter(
                f, fieldnames=column_names, delimiter=delimiter)
            writer.writeheader()
        else:
            writer = csv.writer(f, delimiter=delimiter)
            writer.writerow(column_names)
        writer.writerows(rows)
        # for row in rows:
        #     writer.writerow(row)


def read_csv_by_ID(path, delimiter=","):
    rowlist = read_csv(path, delimiter)
    ID = list(rowlist[0].keys())[0]
    result = {row[ID]:row for row in rowlist}
    return result

def read_csv(path, delimiter=","):
    result=[]
    with open(path, "r", newline="", encoding='UTF-8') as f:
        reader = csv.DictReader(f, delimiter=delimiter)
        for row in reader:
            result.append(row)
    return result
        # for row in rows:
        #     writer.writerow(row)


def cosine_similarity(A, B):
    # define two lists or array

    # compute cosine similarity
    cosine = np.dot(A, B)/(norm(A)*norm(B))
    return cosine


def get_clean_title(title, eID=None, obsidian=False):
    """
    Cleans a given title to make it suitable for filenames.

    Parameters:
    title (string): The title. "episode1"
    eID (int): The position of that episode in the playlist. 

    Returns:
    string: cleaned title for use as filename
    """
    replace_dict = {
        "–": "-",
        "’": "´",
        " ": " ",
    }
    clean_title = title
    for each in noFileChars:
        clean_title = clean_title.replace(each, '')
    for key, value in replace_dict.items():
        clean_title = clean_title.replace(key, value)
    if obsidian:
        clean_title = clean_title.replace("#", '')
    if eID:
        clean_title = fill_digits(eID, 3) + '_' + clean_title
    return clean_title


def get_edited_path(clean_title, input_path):
    if input_path[-1] != '\\':
        input_path += '\\'
    json_path = input_path + EDITFOLDER + '\\' + clean_title + '.json'
    return json_path


def show_newest_files(input_path, files):
    # todo: Test if this could be replaced by:
    # files = [x for x in sort(os.path.getmtime(input_path+'\\'+x))]
    dates = []
    for file in files:
        date = os.path.getmtime(input_path+'\\'+file)
        dates.append(date)
    order = np.argsort(dates)
    files_sorted = []
    for position in reversed(order):
        files_sorted.append(files[position])
    return files_sorted


def clean_dict(dict, playlist_name):
    def default_cleaner(k, v):
        return v

    def split_authors(v, temp):
        temp2 = cut_out(v, dict_author)
        temp['author_type'] = remove_quot(temp2['author_type'])
        temp['author_name'] = remove_quot(temp2['author_name'])
        return temp

    cleaners = {
        'id': default_cleaner,
        'date': time_converter,
        'runtime': time_converter,
        'name': default_cleaner,
        'title': default_cleaner,
        'description': default_cleaner,
        'publisher': default_cleaner,
        'language': default_cleaner,
        '@type': default_cleaner,
        'accessMode': default_cleaner,
        'url': default_cleaner,
        'image': default_cleaner
    }

    temp = {}
    for k, v in dict.items():
        v = remove_quot(v)
        if k in cleaners:
            temp[k] = cleaners[k](k, v)
        elif k == 'author':
            temp[k] = split_authors(v, temp)
        else:
            temp[k] = default_cleaner(k, v)
    return temp


def compare_titles(try_this, comp, playlist_name):
    similarity_score = 0
    for struc in try_this:
        if struc in try_this and struc in comp:
            if comp[struc] == try_this[struc]:
                similarity_score += title_structure[playlist_name][struc][1]
                # We have found the matching episode and it has the ID idx
    return similarity_score


title_structure = {
    'BMZ':   {
        'playlist_name':   [r'(^B+MZ)', 0],
        'void1':   [r'([- ]{0,3})', 0],
        'special':   [r'(\w*[- ]{0,3})', 0.1],
        'void2':   [r'([ # ]{0,3})', 0],
        'eID':   [r'(\d*)', 1],
        'void3':   [r'([: ,]{0,2})', 0],
        'title':   [r'(.*)', 1]
    },
    'Hagrids Hütte':   {
        'eID':   [r'(^[\d\w]{1}\.[\d]{2})', 1],
        'void1':   [r'[- ]{0,3}\w*[- ]{0,3}', 0],
        'title':   [r'.*', 1]
    }
}


def title_mine(title, playlist_name):
    temp = {}
    splits = title_structure[playlist_name]
    for split in splits:
        sep = splits[split][0]
        found = re.split(sep, title, 1)
        if len(found) > 2:
            if 'void' not in split and len(found[1]) > 0:
                temp[split] = found[1]
                # temp[split] = str(found[1])
            title = found[2]
    return temp


def forge_title(title, eID, playlist_name):
    eID += 1  # use ID+1 as eID
    pieces = title_mine(title, playlist_name)
    if 'eID' in pieces:
        eID = pieces['eID']
    elif eID > 128:
        eID += 1
    title = ''
    if 'title' in pieces:
        title = pieces['title']
    title = '#' + str(eID) + ": " + title
    # No playlist_name in title:
    title = re.sub(re.escape(playlist_name) + ' *:*', '', title, 1)
    return title


# Author_name and _type is nested in author, so they have to be extracted as well (for cut_out function)
dict_author = {
    "final_destination_void": "}",
    'author_name': ',"name":',
    'author_type': '{"@type":'
}


def fill_digits(var_s, cap=2):
    """Add digits until cap is reached"""
    var_s = str(var_s)
    while len(var_s) < cap:
        var_s = "0" + var_s
    return var_s


def remove_quot(var_s):
    """Remove leading and closing quotation marks"""
    if var_s:
        if var_s[0] == '"':
            var_s = var_s[1:]
        if var_s[-1] == '"':
            var_s = var_s[:-1]
    return var_s


def get_brand(t):
    if "spotify.com" in t:
        return "Spotify"
    elif "youtube.com" in t:
        return "YouTube"
    elif "bnwiki.de" in t:
        return "BNW"
    return 'Link'


def setup_infos(playlist_info, episodes_info, input_path):
    if not playlist_info:
        with open(input_path + 'playlist_info.json', encoding='utf-8') as json_file:
            # Todo: if an "old replacement"
            playlist_info = json.load(json_file)
    if not episodes_info:
        with open(input_path + 'episodes_info.json', encoding='utf-8') as json_file:
            episodes_info = json.load(json_file)
            episodes_info = {int(k): v for k, v in episodes_info.items()}
    return [playlist_info, episodes_info]


def get_data_folders():
    """Goes up once, then down the data path"""
    my_path = os.getcwd()
    data_path = os.path.dirname(my_path) + '\\data\\'
    playlist_names = [f for f in listdir(
        data_path) if not isfile(join(data_path, f))]
    playlist_names.remove('sample')
    return [data_path, playlist_names]


def cut_hypen(var_s):
    """Cut a string at the first hypen"""
    var_s = var_s.split(" -", 1)[0]
    var_s = var_s.split("-", 1)[0]
    var_s = remove_quot(var_s)
    return var_s


def cut_out(var_s, dict):
    """Cut a string at predetermined marks"""
    storage = {}
    for key, value in dict.items():
        temp = var_s.split(value, 1)
        canidate = ""
        if len(temp) > 1:
            var_s = temp[0]
            canidate = temp[1]
        else:
            canidate = temp[0]
        if 'void' not in canidate and len(canidate) > 0:
            storage[key] = canidate
    return storage


def time_converter(var_s):
    if "Std." in var_s or "Min." in var_s or "Sek." in var_s:
        # Time: HH:MM:SS
        m = re.search(r'\d*(?= *Std.)', var_s)
        std = "" if m is None else m.group(0)
        time = fill_digits(std) + ":"
        m = re.search(r'\d*(?= *Min.)', var_s)
        min = "" if m is None else m.group(0)
        time += fill_digits(min) + ":"
        m = re.search(r'\d*(?= *Sek.)', var_s)
        sek = "" if m is None else m.group(0)
        time += fill_digits(sek)
        return time
    else:
        # Date: YYYY-MM-DD
        list = var_s.split(' ')
        if len(list) == 1:
            # TODO: catch if date is already (somewhat) formatted
            return var_s
        day = ""
        month = list[0]
        year = list[1]
        # See if first entry is really a month
        just_once = 1
        while just_once == 1:
            if month == 'Jan.':
                month = '01'
            elif month == 'Feb.':
                month = '02'
            elif month == 'März':
                month = '03'
            elif month == 'Apr.':
                month = '04'
            elif month == 'Mai':
                month = '05'
            elif month == 'Juni':
                month = '06'
            elif month == 'Juli':
                month = '07'
            elif month == 'Aug.':
                month = '08'
            elif month == 'Sept.':
                month = '09'
            elif month == 'Okt.':
                month = '10'
            elif month == 'Nov.':
                month = '11'
            elif month == 'Dez.':
                month = '12'
            else:
                # If first entry is not a month:
                #   redo check with second entry
                #   and assume year is current year
                just_once += 1
                year = str(datetime.now().year)
                month = list[1]
                day = str(list[0].replace(".", ""))
                day = "-" + fill_digits(day)
            just_once -= 1
        # reminder: day is either "" or "-DD"
        return year + '-' + month + day


def cosine_similarity_text(seq1, seq2):
    # Convert the sequences to word count dictionaries
    seq1_word_count = Counter(seq1)
    seq2_word_count = Counter(seq2)

    # Get the set of unique words in both sequences
    common_words = set(seq1_word_count.keys()) & set(seq2_word_count.keys())

    # If the sequences have no words in common, return 0
    if not common_words:
        return 0

    # Calculate the dot product of the sequences
    dot_product = sum(
        seq1_word_count[word] * seq2_word_count[word] for word in common_words)

    # Calculate the Euclidean norms of the sequences
    seq1_norm = math.sqrt(
        sum(seq1_word_count[word] ** 2 for word in seq1_word_count))
    seq2_norm = math.sqrt(
        sum(seq2_word_count[word] ** 2 for word in seq2_word_count))

    # Return the Cosine similarity
    return dot_product / (seq1_norm * seq2_norm)


def similar(seq1, seq2, method='levenshtein', level=-1):
    """
    Compare two sequences and return True if they are similar enough, based on the chosen method and threshold.

    Parameters
    ----------
    seq1 : str
        The first sequence to compare.
    seq2 : str
        The second sequence to compare.
    method : str, optional
        The method to use for comparing the sequences. Supported values are 'cosine', 'levenshtein', 'jaccard', and 'sequencematcher'.
        Default is 'levenshtein'.
    level : float, optional
        The similarity threshold. Sequences with a similarity score greater than or equal to this value will be considered
        similar. Default is 0.9.

    Returns
    -------
    bool
        True if the sequences are similar enough, False otherwise.

    Raises
    ------
    ValueError
        If the specified method is not supported.
    """
    if level == -1:
        level = SIMILAR_ENOUGH[method]
    if method == 'levenshtein':
        distance = Levenshtein.distance(seq1.lower(), seq2.lower())
        max_length = max(len(seq1), len(seq2))
        return 1 - distance / max_length > level
    # todo: make jaccard worth
    elif method == 'jaccard':
        seq1 = set(re.findall(r'\w+', seq1.lower()))
        seq2 = set(re.findall(r'\w+', seq2.lower()))
        intersection = seq1 & seq2
        union = seq1 | seq2
        return len(intersection) / len(union) > level
    elif method == 'sequencematcher':
        return difflib.SequenceMatcher(a=seq1.lower(), b=seq2.lower()).ratio() > level
    elif method == 'cosine':
        return cosine_similarity_text(seq1, seq2) > level
    else:
        raise ValueError('Invalid method: {}'.format(method))


def extract_file_folder(path: str) -> tuple:
    """Return a tuple containing two lists: one for the file names and one for the folder names in the path."""
    file_names = [f for f in listdir(path) if isfile(join(path, f))]
    folder_names = [f for f in listdir(path) if not isfile(join(path, f))]
    return file_names, folder_names


def concatenate_nested_files(folders_in, root_path):
    files = []
    for folder in folders_in:
        if folder[0] == '.':
            continue
        new_path = root_path + "\\" + folder
        data_files, folders = extract_file_folder(new_path)
        data_files += concatenate_nested_files(folders, new_path)
        for file in data_files:
            files.append(folder + "\\" + file)
    return files


def json2dict(dict_var, name, path=os.getcwd()):
    json_pos = name.find(".json")
    if json_pos != len(name)-5:
        name += '.json'
    if path[-2:] != "\\":
        path += "\\"
    filepath = path + name
    with open(filepath, encoding='utf-8') as json_file:
        dict_var.update(json.load(json_file))
    return os.path.getmtime(filepath)


def lemmatize(var_dict, nlp):
    # this might change the reslts
    wordcount = len(var_dict)
    print('lemmatizing ' + str(wordcount) + ' words', flush=True)
    lemma_dict = {}
    lemmacon = {}
    # this currently turns 500kg to 500 kg, makes it 2 lemmata and thus brakes the sync
    # lemmas = nlp(' '.join(list(var_dict.keys())))
    min = 0
    steps = 1000
    words = list(var_dict.keys())
    lemmas_total = []
    for i in range(int(wordcount/1000)+1):
        words_here = words[i*steps:(i+1)*steps]
        lemmas = nlp(' '.join(words_here))
        if len(words_here) != len(lemmas):
            lemmas = []
            for word in words_here:
                check_lemmas = nlp(word)
                check_lemma_ = []

                for l in check_lemmas:
                    check_lemma_.append(l.lemma_)
                if len(check_lemma_) > 1:
                    print(
                        f"{word} gets lemmatized to {'; '.join(check_lemma_)}", flush=True)
                lemma = "".join(check_lemma_)
                lemmas.append(lemma)
        lemmas_total += lemmas

    for idx, [word, value] in enumerate(var_dict.items()):
        # check_lemmas = nlp(word)
        # check_lemma_ = []
        # if idx % 1000 == 0:
        # print(str(idx), flush=True)
        # for l in check_lemmas:
        # check_lemma_.append(l.lemma_)
        # if len(check_lemmas) > 1:
        # lemma = nlp(word)[0]
        # lemma = lemmas[idx].lemma_
        # if lemma != check_lemma_[0]:
        #     if lemma == check_lemma_[0].replace("ß", "ss"):
        #         continue
        #     print(f"{word} gets lemmatized to {lemma} and {check_lemma_}")
        # TODO: Find out why lemma is sometimes not automatically lower case
        lemma = lemmas_total[idx]
        if type(lemma) is not str:
            lemma = lemma.lemma_
        lemma = lemma.lower()
        # Todo: Not sure how to handle "Alte Rechtschreibung" yet
        # lemma = lemma.replace("ß", "ss"),
        # Due to lexicon being sorted, lemma elements should be automatically sorted
        # sort = False
        # if lemma in lemma_dict:
        #     sort = True
        lemma_dict[word] = lemma
        nested_add(lemmacon, [lemma], {word: value})
        # if sort:
        #     lemma_dict[lemma] = {k: v for k, v in sorted(
        #         lemma_dict[lemma].items(), reverse=True, key=lambda item: item[1])}
    print('reduced to  ' + str(len(lemmacon)) + ' words', flush=True)
    return lemmacon, lemma_dict


def get_transcript(input_path, filename):
    path_json = input_path + AUDIOFOLDER + '\\' + filename
    with open(path_json, encoding='utf-8') as json_file:
        transcript = json.load(json_file)
    return transcript


def dict2json(dict_var, name, path=os.getcwd()):
    json_pos = name.find(".json")
    if json_pos != len(name)-5:
        name += '.json'
    if path[-2:] != "\\":
        path += "\\"
    json_object = json.dumps(dict_var, indent=4, ensure_ascii=False)
    with open(path + name, "w", encoding='utf8') as outfile:
        outfile.write(json_object)

# dict stuff:
# https://stackoverflow.com/questions/14692690/access-nested-dictionary-items-via-a-list-of-keys


def nested_set(dic, keys, value):
    for key in keys[:-1]:
        if key in dic and dic[key] == None:
            dic[key] = {}
        dic = dic.setdefault(key, {})
    dic[keys[-1]] = value


def nested_add(dic, keys, value):
    for key in keys[:-1]:
        if key in dic and dic[key] == None:
            dic[key] = {}
        dic = dic.setdefault(key, {})
    if keys[-1] in dic:
        if type(value) == dict:
            for key_i, value_i in value.items():
                nested_add(dic, keys + [key_i], value_i)
        else:
            dic[keys[-1]] += value
    else:
        dic[keys[-1]] = value


def nested_get(dic, keys):
    for key in keys:
        dic = dic[key]
    return dic


def split_layers(string, characters=" ", keep_char=False, layer=0):
    if layer == len(characters):
        return [string]
    x = characters[layer]
    pieces = []
    if x in string:
        layer_pieces = string.split(x)
        for piece in layer_pieces:
            if piece == '':
                continue
            inner_results = split_layers(piece, characters, keep_char, 1+layer)
            if not inner_results:
                continue
            for s in inner_results[:-1]:
                pieces.append(s)
            last_piece = inner_results[-1]
            if keep_char:
                last_piece += x
            pieces.append(last_piece)
    else:
        pieces = split_layers(string, characters, keep_char, 1+layer)
    return pieces


def wordify(string):
    regex = r'\b[\w]+(?<!\d)-*\w*\b'
    res = [x.lower() for x in re.findall(regex, string)]
    return res


def sentencify(string):
    return split_layers(string, '.!?', True)



import os
from ..core import helper
import bibtexparser

import csv
import json
import re

# Todo: make this work directly from API / Website
# import urllib.request
# page = urllib.request.urlopen('https://orkg.org/api/classes/ResearchField/resources/?page=0&size=999')
# print(page.read())
# a = page.read()
# RF_WEB = json.loads(page.text)

# Directly from dictionary
with open(os.path.join('ORKG', 'ResearchFields.json')) as json_file:
    data = json.load(json_file)

RF_map = {}
for rf in list(data.values())[0]:
    RF_map.update({rf['id']: rf['label']})
RF_map = {k: v for k, v in sorted(
    RF_map.items(), key=lambda item: int(item[0][1:]))}

with open(os.path.join('ORKG', 'RF_Map.json'), 'w') as outfile:
    json.dump(RF_map, outfile, indent=4)

RF_patterns = {}
for raw_field in RF_map.values():
    pot_fields = raw_field
    if " " not in raw_field:
        continue
    fields = []
    master = ""
    patterns = []
    if " of " in raw_field:
        master = raw_field.split(" of ", 1)[0]
        pot_fields.replace(f"{master} of ", "")
        master = re.escape(master) + r"*"
    elif " and " in raw_field:
        temp = raw_field.split(" and ", 1)[1]
        if " " in temp:
            temp = " " .join(temp.split(" ", 1)[1:])
            master = re.escape(temp) + r"*"
    fields = re.split(',|/| and ', raw_field)
    fields = [field.strip(' ') for field in fields]
    if "" in fields:
        fields.remove("")
    if len(fields) > 1:
        for field in fields:
            pat = r"\b" + re.escape(field) + r"\b"
            patterns.append(pat + r".*" + master)
            patterns.append(master + r".*" + pat)
    if patterns:
        RF_patterns[raw_field] = patterns


def update_string(data, key, value):
    data.update({f"paper:{key}": value})
    return data


def update_list(data, key, value):
    values = value.split(", ")
    data.update({f"paper:{key}": ";".join(values)})
    return data


def update_authors(data, key, value):
    values = value.split(" and ")
    true_values = []
    for value in values:
        true_values += value.split(" and ")
    values = true_values
    true_values = []
    for value in values:
        result = value
        if "," in value:
            group = value.split(",")
            if len(group) == 2:
                result = group[1] + " " + group[0]
            else:
                print('Found an author with >2 "," : ' + value)
        true_values.append(result)
    values = true_values
    data.update({f"paper:{key}": ";".join(values)})
    return data


def update_date(data, key, value):
    values = value.split("-")
    data.update({f"paper:publication_year": values[0]})
    if len(values) > 1:
        data.update({f"paper:publication_month": values[1]})
    return data


def identify_field_weight(field, content):
    if field in content:
        return 1
    elif field in RF_patterns:
        # Is: "Electrical Engineering"
        # Field: "Electrical and Computer Engineering"
        for pattern in RF_patterns[field]:
            if re.search(pattern, content):
                return 1  # maybe weigh this less?


def update_RF(data, key, tex):
    # TODO
    # Attempt to get RF:
    weights = {
        "booktitle": 10,
        "journaltitle": 10,
        "title": 4,
        # "abstract": 1,
    }
    assume = {}
    for search_key, weight in weights.items():
        if search_key in tex:
            for RF_id, field in RF_map.items():
                # TODO:
                value = identify_field_weight(field, tex[search_key])
                if value:
                    if RF_id not in assume:
                        assume[RF_id] = 0
                    assume[RF_id] = value * weight
    assume = {k: v for k, v in sorted(
        assume.items(), key=lambda item: item[1], reverse=True)}
    highest = 0
    results = []
    for RF_id, value in assume.items():
        # TODO multiple matches?

        for result in results:
            # check if "Engineering" could be made "Computer Engineering"
            comp = identify_field_weight(RF_map[result], RF_map[RF_id])
            if comp:
                # "Engineering" is in "Computer Engineering"
                results.remove(result)
                results.append(RF_id)
                highest = value
                continue

        if value >= highest:
            if RF_id not in results:
                results.append(RF_id)
            highest = value

            # 'R137': 10,     "R137": "Numerical Analysis/Scientific Computing",
            # 'R194': 10,     "R194": "Engineering",
            # 'R237': 10,     "R237": "Electrical and Computer Engineering",
            # 'R241': 10,     "R241": "Electrical and Electronics",
            # 'R254': 10,     "R254": "Materials Science and Engineering",
            # 'R201': 4     "R201": "Structures and Materials",
    # If you want to be able to read the results
    # results = [RF_map[x] for x in results]
    if results:
        # data.update({f"paper:{key}": "orkg:" + "; orkg:".join(results)})
        # data.update({f"paper:{key}": "; ".join(results)})
        # data.update({f"paper:{key}" : f"orkg:{results[-1]}"})
        data.update({f"paper:{key}": f"{results[-1]}"})
    return data


def update_research_problem(data, key, value):
    # TODO
    # values = value.split(" and ")
    # [
    #     "This work addresses"
    # ]
    # data.update({f"paper:{key}": "; ".join(values)})
    return data


def tex_to_csv_format(tex):
    data = {}
    handlers = {
        "title": update_string,
        "authors": update_authors,
        "publication_year": update_date,
        "published_in": update_string,
        "research_field": update_RF,
        "doi": update_string,
        "url": update_string,
        "research_problem": update_research_problem,

        # End of fixed:
        "keywords": update_list,
        # "location": update_string,
        # "abstract": update_string,
        # "publication_month": update_string,
    }
    tex_map = {
        "authors": ["author"],
        "publication_year": ["date"],
        "published_in": ["booktitle", "journaltitle"],
    }
    tex_advanced = [
        "research_field",
        "research_problem",
    ]
    for key in handlers:
        searchfor = [key]
        if key in tex_map:
            searchfor += tex_map[key]
        for search_key in searchfor:
            if search_key in tex:
                data = handlers[key](data, key, tex[search_key])
            elif search_key in tex_advanced:
                data = handlers[key](data, key, tex)
    return data

def filter_by_k_score(bib_database, bib_path, threshold = None):
    k_score_file = os.path.join(os.path.dirname(bib_path), "k_score.csv")
    try:
        scores = helper.read_csv_by_ID(k_score_file)
    except Exception as e:
        print(e)
        print("No k_scores.csv found. Using full BibTeX")
        return bib_database.entries

    result = []
    bib_db_by_ID = [e['ID'] for e in bib_database.entries]
    for name, entry in scores.items():
        if threshold and float(entry['total_score']) < threshold:
            continue 
        if name in bib_db_by_ID:
            # TODO: Chance to append the score to the entry here.
            result.append(bib_database.entries[bib_db_by_ID.index(name)])
    if not result:
        print("No matches with k_score.csv found. Using full BibTeX")
        return bib_database.entries
    return result


def main(bib_path, csv_path=None, threshold=None):
    with open(bib_path, encoding='UTF8') as bibtex_file:
        bib_database = bibtexparser.load(bibtex_file)
    orkg_data_csv = []
    header_names = set()
    bib_database_entries = filter_by_k_score(bib_database, bib_path, threshold)
    for idx, entry in enumerate(bib_database_entries):
        orkg_data_csv.append(tex_to_csv_format(entry))
        header_names.update(orkg_data_csv[-1].keys())
        if idx % 50 == 0:
            print(f"{idx}/{len(bib_database_entries)} done")

    if not csv_path:
        csv_path = os.path.join(os.path.dirname(bib_path), "ORKG.csv")
    helper.write_to_csv(header_names, orkg_data_csv, csv_path)



from ...core import helper
from ...ORKG import bibtex_to_csv as ORKG_csv
from ...review import k_score
#TODO: this file no longer exists. Need to be re-checked before this code is revived.
from ...extract import util_bibtex

import re
import os
from typing import List
from typing import Dict
from os import listdir
from os.path import isfile, join
import markdown

#TODO This is an old import, is now part of config
# from ...config import personal


LITERATURE = "Reading notes"


def extract_field_values(input_string: str) -> dict:
    """Extract values of fields from input string using regular expressions."""
    date = re.search(r'(?<=year: ).*?(?=\n)', input_string).group()
    title = re.search(r'(?<=title: ).*?(?=\n)', input_string).group()
    authors = re.search(r'(?<=authors: ).*?(?=\n)', input_string).group()
    in_ = re.search(r'(?<=in: ).*?(?=\n)', input_string).group()
    page = re.search(r'(?<=page: ).*?(?=\n)', input_string).group()
    cite_title = re.search(
        r'(?<=citeTitle:[\n ]).*?(?=\n)', input_string).group()
    cite_id = re.search(r'(?<=citeID:[\n ]).*?(?=\n)', input_string).group()
    url = re.search(r'(?<=url: ).*?(?=\n)', input_string).group()
    zotero = re.search(r'(?<=zotero: ).*?(?=\n)', input_string).group()
    tags = re.search(r'(?<=tags:: ).*?(?=\n)', input_string).group()
    read = re.search(r'(?<=read\?:: ).*?(?=\n)', input_string).group()
    comment = re.search(r'(?<=comment:: ).*?(?=\n)', input_string).group()
    return {
        "date": date,
        "title": title,
        "authors": authors,
        "in_": in_,
        "page": page,
        "cite_title": cite_title,
        "cite_id": cite_id,
        "url": url,
        "zotero": zotero,
        "tags": tags,
        "read": read,
        "comment": comment
    }


def extract_pdfs(outer_path, files):
    pdfs = []
    if files:
        for file_i in files.values():
            if file_i.endswith(".pdf"):
                # to remove \ in filenames
                file_i = file_i.replace("\\", "")
                pdfs.append(os.path.join(
                    outer_path, file_i).replace("/", "\\"))
    return pdfs


def build_literature_note(field_values: dict, rest: str, lib_path: str, cfg) -> str:
    """Build output string using string formatting method and join()."""
    # YAML
    date = field_values.get('date', '')
    title = field_values.get('title', '')
    author = field_values.get('author', '')
    in_ = field_values.get('in_', '')
    in_abrev = field_values.get('in_', '')
    pages = field_values.get('pages', '')
    cite_title = field_values.get('cite_title', '')
    cite_id = field_values.get('cite_id', '')
    url = field_values.get('url', '')
    zotero = field_values.get('zotero', '')
    tags = field_values.get('tags', '')
    read = field_values.get('read', '')
    comment = field_values.get('comment', '')

    learning = field_values.get('learning', '')
    keyword_links = field_values.get('keyword_links', '')
    key_notes = field_values.get('key_notes', '')
    quotes = field_values.get('quotes', '')

    editorbtype = field_values.get('editorbtype', '')
    editorb = field_values.get('editorb', '')
    eventtitle = field_values.get('eventtitle', '')
    pagetotal = field_values.get('pagetotal', '')
    pmcid = field_values.get('pmcid', '')

    langid = field_values.get('langid', '')
    editor = field_values.get('editor', '')
    publisher = field_values.get('publisher', '')
    booktitle = field_values.get('booktitle', '')
    journaltitle = field_values.get('journaltitle', '')
    shortjournal = field_values.get('shortjournal', '')
    url = field_values.get('url', '')
    urldate = field_values.get('urldate', '')
    abstract = field_values.get('abstract', '')
    series = field_values.get('series', '')
    volume = field_values.get('volume', '')
    doi = field_values.get('doi', '')
    isbn = field_values.get('isbn', '')
    issn = field_values.get('issn', '')
    number = field_values.get('number', '')
    rights = field_values.get('rights', '')
    keywords = field_values.get('keywords', '')
    location = field_values.get('location', '')
    ENTRYTYPE = field_values.get('ENTRYTYPE', '')
    ID = field_values.get('ID', '')
    file = field_values.get('file', '')
    shorttitle = field_values.get('shorttitle', '')
    pmid = field_values.get('pmid', '')
    note = field_values.get('note', '')
    eprint = field_values.get('eprint', '')
    eprinttype = field_values.get('eprinttype', '')
    issue = field_values.get('issue', '')

    if not in_:
        if booktitle:
            in_ = booktitle
        elif journaltitle:
            in_ = journaltitle
    if not in_abrev:
        if shortjournal:
            in_abrev: shortjournal

    if not cite_title:
        cite_title = f"[[note_path/@{ID}|{title}]]"
    if not cite_id:
        cite_id = f"\[[[note_path/@{ID}|00]]\]"
    if not zotero:
        temp_zot = "zotero://select/items/@" + ID
        zotero = f"[{temp_zot}]({temp_zot})"
    temp_tags = ["#science ", "#literature "]
    if tags:
        if not tags.endswith(" "):
            tags += " "
    for temp_tag in temp_tags:
        if temp_tag not in tags:
            tags += temp_tag
    bib_path = os.path.join(cfg['vault_path'], lib_path).replace("\\", "/")
    pdf = ""
    pdfs = extract_pdfs(bib_path, file)
    for pdf_ in pdfs:
        pdf += f'[Acrobat]({pdf_.replace(" ", "%20")})\n'
    lib_path = lib_path.replace("\\", "/")
    if not pdf:
        temp_tags = ["#missingPDF "]
        for temp_tag in temp_tags:
            if temp_tag not in tags:
                tags += temp_tag
    if not read:
        read = "#read/false "

    # todo: make all lower case
    if keywords:
        temp = re.split(", *", keywords)
        if keyword_links:
            keyword_links += " "
        missing = []
        for key in temp:
            search = "[[" + key + "]]"
            if search not in temp:
                missing.append(search)
        keyword_links = " ".join(missing)

    yaml = f"""---
date: {date}
title: \"{title}\"
author: \"{author}\"
editor: \"{editor}\"
publisher: \"{publisher}\"
pages: \"{pages}\"
language: \"{langid}\"
location: \"{location}\"
ENTRYTYPE: \"{ENTRYTYPE}\"
in: \"{in_}\"
short: \"{in_abrev}\"
series: \"{series}\"
volume: {volume}
doi: {doi}
isbn: {isbn}
url: \"{url}\"
ID: \"{ID}\"
issn: {issn}
number: {number}
rights: \"{rights}\"
keywords: \"{keywords}\"
urldate: {urldate}
shorttitle: \"{shorttitle}\"
pmid: {pmid}
note: \"{note}\"
eprint: \"{eprint}\"
eprinttype: \"{eprinttype}\"
issue: \"{issue}\"
editorbtype: \"{editorbtype}\"
editorb: \"{editorb}\"
eventtitle: \"{eventtitle}\"
pagetotal: {pagetotal}
pmcid: {pmcid}
---
""".replace(': ""', ": ")

    infos = f"""```citeTitle:
{cite_title}
```
```citeID:
{cite_id}
```
---
url: {url}
zotero: {zotero}
{pdf}
---
tags:: {tags}
read:: {read}
comment:: {comment}

---
"""
    text = f"""# Learning
{learning}

## Keywords
{keyword_links}

## Key notes
{key_notes}

## Quotes
{quotes}

# Abstract
{abstract}
"""
    for name, path in dict(file).items():
        path = path.replace("\\", "")
        # if include files to yaml
        # if name.startswith("path_"):
        #     name = name.replace("path_", "", 1)
        #     name = name.replace("_", " ")
        text += f"""
# {name}
[Acrobat]({os.path.join(bib_path, path).replace(" ", "%20")})
[[{lib_path}/{path}|{path.split("/", 2)[2]}]]
![[{lib_path}/{path}]]
"""

    # todo: make sure rest is never needed anymore
    return yaml + infos + text + rest


def reformat(input_string: str, splitter: str = "\n# Learning\n") -> tuple:
    pieces = input_string.split(splitter, 1)
    work = pieces[0]
    rest = pieces[1]

    field_values = extract_field_values(work)
    output_string = build_literature_note(field_values, splitter + rest)
    return output_string, field_values


def categorize_files(data_files: List[str], cfg: Dict) -> Dict[str, List[str]]:
    """Sorts files into notes, zotero, literature_notes and others"""
    files = {
        "notes": [],
        "zotero": [],
        "literature_notes": [],
        "others": [],
    }
    for file in data_files:
        if file.endswith('.md'):
            if file.startswith(cfg["literature_notes_folder_name"] + '\\'):
                files["literature_notes"].append(file)
            else:
                files["notes"].append(file)
        elif file.startswith(cfg["zotero_folder_name"] + '\\'):
            files["zotero"].append(file)
        else:
            files["others"].append(file)
    return files


def reformat_literature_files(cfg, files: Dict) -> Dict[str, str]:
    bib = {}
    for lit in files["literature_notes"]:
        filepath = os.path.join(cfg['vault_path'], lit)
        with open(filepath, 'r', encoding='UTF8') as f:
            text = f.read()
        text, var_dict = reformat(text)
        with open(filepath, 'w', encoding='UTF8') as f:
            f.write(text)
        pieces = var_dict["cite_title"].split('|', 1)
        uid = pieces[0]
        title = pieces[1]
        bib.update({uid: title})
    return bib


def format_title(title: str) -> List[str]:
    """Format title for matching by replacing '?', ':', and '–' with alternate characters."""
    check = [title]
    if "?" in title:
        temp = [token.replace("?", "❓") for token in check]
        check += temp
    if ":" in title:
        temp = [token.replace(":", ";") for token in check]
        check += temp
    if "–" in title:
        temp = [token.replace("–", "-") for token in check]
        check += temp
    return check


def reformat_note_files(cfg, notes: List[str], bib: Dict[str, str]) -> None:
    for note in notes:
        filepath = cfg['vault_path'] + '\\' + note
        with open(filepath, 'r', encoding='UTF8') as f:
            text = f.read()
        text_in = text
        for uid, title in bib.items():
            check = format_title(title)
            for token in check:
                text = re.sub(re.escape("[[{}]]".format(token)),
                              "[[{}|{}]]".format(uid, title), text, flags=re.I)
                text = re.sub(re.escape("[[{}".format(token)),
                              "[[{}".format(uid), text, flags=re.I)
        if text != text_in:
            with open(filepath, 'w', encoding='UTF8') as f:
                f.write(text)


def read_from_note(var_dict, note):
    patterns = [
        (r'(?<=\ntags:: ).*(?=\nread:: )', 'tags'),
        (r'(?<=\nread:: ).*(?=\ncomment:: )', 'read'),
        (r'(?<=\ncomment:: ).*(?=\n---\n)', 'comment'),
        (r'(?<=\n# Learning\n).*(?=\n## Keywords\n)', 'learning'),
        (r'(?<=\n## Keywords\n).*(?=\n## Key notes\n)', 'keyword_links'),
        (r'(?<=\n## Key notes\n).*(?=\n## Quotes\n)', 'key_notes'),
        (r'(?<=\n## Quotes\n).*(?=\n# Abstract\n)', 'quotes')
    ]
    for pattern, key in patterns:
        m = re.search(pattern, note, flags=re.DOTALL)
        if m:
            var_dict[key] = m.group().rstrip()
    return var_dict


def create_zotero_notes(files, cfg):
    # import bib
    libname = "Meine Bibliothek"
    lib_path = f"{cfg['zotero_folder_name']}\\{libname}"
    bib_name = f"{cfg['vault_path']}\\{cfg['zotero_folder_name']}\\{libname}\\{libname}.bib"
    bib_db_clean = util_bibtex.clean(bib_name)

    if cfg['DO_SCORE']:
        k_score_dict = {}
    for ID, entry_dict in bib_db_clean.items():
        rest = ""
        if cfg['DO_OBSIDIAN']:
            filepath = os.path.join(
                cfg['vault_path'], cfg['literature_notes_folder_name'] + "\\@" + ID + ".md")
            # todo: handle if ReadNotes does exist (somewhat handled)
            if os.path.isfile(filepath):
                with open(filepath, 'r', encoding='UTF8') as f:
                    temp_note = f.read()
                    entry_dict = read_from_note(entry_dict, temp_note)

        if 'file' in entry_dict:
            pdfs = extract_pdfs(lib_path, entry_dict['file'])
            if pdfs:
                print("Working on " + ID)

                for file in pdfs:
                    entry_dict = k_score.read_pdf(
                        os.path.join(cfg['vault_path'], file), entry_dict, cfg)
        note = build_literature_note(
            entry_dict, rest, lib_path, cfg)
        if cfg['DO_OBSIDIAN'] and temp_note != note:
            with open(filepath, 'w', encoding='UTF8') as f:
                f.write(note)
    # TODO: Check if exists and needs to be reformated:
    # for now: just overwrite
    # bib = reformat_literature_files(cfg['vault_path'], files)

    # for file in cfg['literature_notes_folder_name']:
    # file_prefix = f"{cfg["zotero_folder_name"]}\\{libname}\\files"
    # for filename in files["zotero"]:
    #     if filename.startswith(file_prefix):
    #         pass
        #         # here lies the attatchment
        if pdfs:
            csv_path = os.path.join(cfg['vault_path'], lib_path)
            if "k_score" in entry_dict:
                k_score_dict[entry_dict['ID']] = entry_dict['k_score']
                k_score.write_to_file(k_score_dict, csv_path, cfg)

    return files


def check_cfg(cfg, personal):
    def merge(cfg):
        return os.path.join(cfg["obsidian_path"], cfg["vault_name"])
    assume = {
        "obsidian_path": personal.OBSIDIAN_PATH,
        "vault_name": "TIB_Obsidian",
        "vault_path": merge,
        "zotero_folder_name": "Zotero",
        "literature_notes_folder_name": "Reading notes",
        "DO_OBSIDIAN": False,
        "DO_SCORE": False,
    }
    for key, value in assume.items():
        if key not in cfg:
            if callable(value):
                cfg[key] = value(cfg)
            else:
                cfg[key] = value
    return cfg

def worth_it(cfg):
    worth_it = False
    for key, value in cfg.items():
        if key.startswith("DO_") and value:
            worth_it = True
    return worth_it

def main(cfg={}):
    # todo: have an option to make this work without obsidian
    cfg = check_cfg(cfg)
    if not worth_it(cfg):
        print("Please select at least one DO_X: TRUE in your config")
        return
    vault_files, vault_folders = helper.extract_file_folder(cfg['vault_path'])
    vault_files += helper.concatenate_nested_files(vault_folders, cfg['vault_path'])
    files = categorize_files(
        vault_files, cfg)
    files = create_zotero_notes(
        files, cfg)

if __name__ == "__main__":
    main()



import os
import subprocess

# import graphviz

def generate_dot_file(node_prep, edge_prep, filename, path, image_format = "png"):
        # Assemble the Graphviz DOT file
    nodes = []
    for node in node_prep:
        ID = node[0]
        if len(node) == 1:
            label = node[0]
        else:
            label = node[1]
        nodes.append(f"{ID} [label=\"{label}\"]")
    edges = []
    for node in edge_prep:
        fID = node[0]
        tID = node[1]
        if len(node) == 2:
            label = ""
        else:
            label = f"[label=\"{node[2]}\"]"
        edges.append(f"{fID} -> {tID}{label}")

    dot_file = "digraph {\n"
    dot_file += "node [shape=ellipse, style=filled, color=lightblue, fontcolor=black, fontsize=12]\n"
    dot_file += "\n".join(nodes)
    dot_file += "\n"
    # dot_file += "edge [dir=none, color=gray, fontcolor=black, fontsize=10]\n"
    dot_file += "edge [color=gray, fontcolor=black, fontsize=10]\n"
    dot_file += "\n".join(edges)
    dot_file += "\n}"

    os.makedirs(path,exist_ok=True)
    filepath = os.path.join(path.replace("\\edited", "\\graph"), filename)
    dot_filename = filepath + ".dot"
    with open(dot_filename, "w", encoding="utf8") as f:
        f.write(dot_file)

    image_filename = filepath + "." + image_format
    # subprocess.run(["dot", "-Kneato", "-v",  "-T" + image_format, "-Gsize=100,100", "-Gdpi=300",  dot_filename, "-o", image_filename.replace("."+image_format, "_neato2."+image_format)])
    # subprocess.run(["dot", "-Kneato", "-v",  "-T" + image_format, dot_filename, "-o", image_filename.replace("."+image_format, "_neato."+image_format)])
    subprocess.run(["dot", "-v",  "-Tsvg", dot_filename, "-o", image_filename.replace("."+image_format, ".svg")])
    subprocess.run(["dot", "-v",  "-T" + image_format, dot_filename, "-o", image_filename])
    # subprocess.run(["dot", "-v",  "-T" + image_format, "-Gsize=50,50", "-Gdpi=300", dot_filename, "-o", image_filename.replace("."+image_format, "2."+image_format)])
    # subprocess.run(["dot", "-Kfdp", "-v",  "-T" + image_format, "-Gsize=50,50", "-Gdpi=300", dot_filename, "-o", image_filename.replace("."+image_format, "_fpd2."+image_format)])
    subprocess.run(["dot", "-Kfdp", "-v",  "-Tsvg", dot_filename, "-o", image_filename.replace("."+image_format, "_fpd.svg")])
    subprocess.run(["dot", "-Kfdp", "-v",  "-T" + image_format, dot_filename, "-o", image_filename.replace("."+image_format, "_fpd."+image_format)])
    return dot_file

def episodes_to_dot_file(episodes, filename, path, image_format = "png"):
    # Create a list of nodes, where each node represents an episode
    
    nodes = []
    for episode_id, ep in episodes.items():
        node_id = f"episode_{episode_id}"
        title = ep['title'].replace('"', "'")
        nodes.append(f"{node_id} [label=\"{title}\"]")

    # Create a list of edges, where each edge represents a common word between two episodes
    edges = []
    for episode_id_1, ep_1 in episodes.items():
        for episode_id_2, ep_2 in episodes.items():
            if episode_id_1 >= episode_id_2:
                continue

            common_words = set(ep_1['words']) & set(ep_2['words'])
            if common_words:
                node_id_1 = f"episode_{episode_id_1}"
                node_id_2 = f"episode_{episode_id_2}"
                edges.append(f"{node_id_1} -> {node_id_2} [label=\"{', '.join(common_words)}\"]")
    dot_file = generate_dot_file(nodes, edges, filename, path, image_format)
    return dot_file



import os
import sys
import json
import warnings

try:
    from pytube import Channel
    from pytube import Playlist
    from pytube import YouTube
except ImportError:
    warnings.warn("Not using YouTube. If you want to use it, run 'pip install bnw_tools[YouTube]'")
    


OVERWRITE = True

noFileChars = '":\<>*?/|,#.\'$^;~%'


def subtitle_clean(subtitle_text):
    subtitle_text = subtitle_text.split('\n\n')
    sent_num = 1

    result = ""
    for i in range((len(subtitle_text) - 1) // 2):
        _, temp_time, temp_text = subtitle_text[i * 2].split('\n')
        temp_start = temp_time.strip().split(' --> ')[0]
        _, next_time, _ = subtitle_text[(i + 1) * 2].split('\n')
        temp_end = next_time.strip().split(' --> ')[0]
        result += str(sent_num) + '\n'
        result += temp_start + ' --> ' + temp_end + '\n'
        result += temp_text.strip() + '\n'
        result += '\n'
        sent_num += 1

    rest = subtitle_text[-1].split('\n')
    _ = rest[0]
    temp_time = ""
    if len(rest) > 1:
        temp_time = rest[1]
    temp_text = ""
    if len(rest) > 2:
        temp_text = rest[2]

    result += str(sent_num) + '\n'
    result += temp_time + '\n'
    result += temp_text.strip() + '\n'
    return result

# Infos


class YouTubeInfo:
    def __init__(self, path: str = None, json_data: dict[str, str] = None):
        if path:
            if os.path.exists(path):
                if not json_data:
                    json_data = {}
                with open(path, "r", encoding="utf-8") as json_file:
                    json_data.update(json.load(json_file))

        self.video_id = json_data['videoDetails']['videoId'] if json_data else None
        self.title = json_data['videoDetails']['title'] if json_data else None
        self.length_seconds = json_data['videoDetails']['lengthSeconds'] if json_data else None
        self.channel_id = json_data['videoDetails']['channelId'] if json_data else None
        self.short_description = json_data['videoDetails']['shortDescription'] if json_data else None
        self.view_count = json_data['videoDetails']['viewCount'] if json_data else None
        self.author = json_data['videoDetails']['author'] if json_data else None
        self.publish_date = json_data['videoDetails']['publish_date'] if json_data else None

        self.cap_codes = json_data['cap_codes'] if json_data and "cap_codes" in json_data else {
        }

        self.thumbnail_urls = []
        if json_data and 'videoDetails' in json_data and 'thumbnail' in json_data['videoDetails']:
            for thumbnail in json_data['videoDetails']['thumbnail']['thumbnails']:
                self.thumbnail_urls.append(thumbnail['url'])

    def get_transcript(self):
        score = {
            "": 0,
            "a.de": 1,
            "a.en": 2,
            "de": 3,
            "en": 4,
        }

        def get_accepted(candidate):
            candidate = candidate.replace("a.", "")
            if len(candidate) > 2:
                if candidate[:2] in score:
                    return candidate[:2]
            return candidate
        best = ["", 0]

        for candidate in self.cap_codes.keys():
            c_score = 1
            key = candidate
            candidate = get_accepted(candidate)
            if key in score:
                c_score = score[key]
            elif candidate in score:
                c_score = score[candidate]
            # Todo: support more than german and english
            # elif "a." in candidate:
            #     c_score = c_score / 2
            else:
                continue
            if c_score > best[1]:
                best = [key, c_score]

        res_language = get_accepted(best[0])

        res_text = self.cap_codes[best[0]] if best[0] in self.cap_codes else ""
        if not res_language:
            print("Supported languages: " + ", ".join(score.keys()))
            print("Given languages: " + ", ".join(self.cap_codes.keys()))
        return res_text, res_language


def print_infos(video, infos_dir):
    # TODO: check if file already exists
    filename = video.title
    for char in noFileChars:
        filename = filename.replace(char, "")
    file_path = os.path.join(infos_dir, filename) + ".json"
    if not OVERWRITE and os.path.isfile(file_path):
        return

    # https://github.com/pytube/pytube/issues/1085

    infos = {}
    not_interesting = [
        "responseContext",
        "playabilityStatus",
        "streamingData",
        "playbackTracking",
        "playerConfig",
        "storyboards",
        "trackingParams",
        "attestation",
        "adBreakParams",
        "playerSettingsMenuData",
        "endscreen",
    ]
    for key, value in video.vid_info.items():
        if key in not_interesting:
            continue
        infos[key] = value

    # publish time not known, just date
    # https://github.com/pytube/pytube/issues/1269
    infos['videoDetails'].update(
        {'publish_date': video.publish_date.isoformat()})

    for cap in video.captions:
        code = cap.code
        caption = video.captions[code]
        # Auto generation findable, but not used yet.
        # todo: Idea: Add Whisper transcript here as "w.de" or whatever.
        # auto_generated = False
        # if code.startswith("a."):
        #     auto_generated = True
        try:
            c1 = caption.generate_srt_captions()
            c2 = subtitle_clean(c1)
            if "cap_codes" not in infos:
                infos["cap_codes"] = {}
            infos["cap_codes"][code] = c2
        except Exception as e:
            print(f"\t\tError doing the {code} captions for: " + filename)
            print(e, flush=True)

    with open(file_path, 'w', encoding='utf8') as json_file:
        json.dump(infos, json_file, ensure_ascii=False, indent=4)

    return


def print_audio(video, audio_dir):
    test = video.title
    for char in noFileChars:
        test = test.replace(char, "")
    test += ".mp4"
    if not os.path.exists(os.path.join(audio_dir, test)):
        print("\t" + test)
        video.streams.get_audio_only().download(output_path=audio_dir)

    return

# TODO: Test https://www.youtube.com/user/coldmirror/videos
# TODO: Test https://www.youtube.com/@coldmirror/videos


def print_channel(c, channel_dir):
    filename = c.channel_name
    for char in noFileChars:
        filename = filename.replace(char, "")
    file_path = os.path.join(channel_dir, filename) + ".json"
    not_interesting = [
        "html",
        "ytcfg",
        # "initial_data",
    ]
    infos = {}
    object_methods = []
    object_values = []
    for method_name in dir(c):
        if method_name.startswith("_"):
            continue
        # if method_name.startswith('__'):
        #     continue
        if method_name.endswith("_html"):
            continue
        if method_name in not_interesting:
            continue
        try:
            if callable(getattr(c, method_name)):
                object_methods.append(method_name)
            else:
                object_values.append(method_name)
        except:
            pass
    for method_name in object_values:
        v = getattr(c, method_name)
        if method_name == "initial_data":
            v = v['metadata']
            del v['channelMetadataRenderer']['availableCountryCodes']
            method_name = "metadata"
        try:
            json.dumps(v)
        except:
            continue
        infos[method_name] = v
    infos["video_count"] = len(c.video_urls._elements)
    infos["video_urls"] = c.video_urls._elements
    # for idx, key in enumerate(c):
    #     try:
    #         infos[idx] = key
    #     except:
    #         pass
    with open(file_path, 'w', encoding='utf8') as json_file:
        json.dump(infos, json_file, ensure_ascii=False, indent=4)


def do_channels(root_path, channel_id, channel_name):

    # Check whether the specified path exists or not
    channel_dir = os.path.join(root_path, channel_name)
    audio_dir_old = os.path.join(channel_dir, "audio")
    audio_dir = os.path.join(channel_dir, "00_audios")
    if os.path.exists(audio_dir_old):
        os.rename(audio_dir_old, audio_dir)

    infos_dir_old = os.path.join(channel_dir, "infos")
    infos_dir = os.path.join(channel_dir, "00_infos")
    if os.path.exists(infos_dir_old):
        os.rename(infos_dir_old, infos_dir)

    paths = [channel_dir, audio_dir, infos_dir]
    for path_ in paths:
        if not os.path.exists(path_):
            os.makedirs(path_)
    os.chdir(channel_dir)

    try_this = [
        f"c/{channel_name}",
        f"channel/{channel_id}",
        f"u/{channel_name}",
        f"user/{channel_id}",

        f"c/{channel_id}",
        f"channel/{channel_name}",
        f"u/{channel_id}",
        f"user/{channel_name}",
        f"@{channel_id}",
        f"@{channel_name}"
    ]

    found = ""
    for this in try_this:
        try:
            # link = f'https://www.youtube.com/{this}/videos'
            link = f'https://www.youtube.com/{this}'
            c = Channel(link)
            # c = Channel(this)
            # c = Channel('https://www.youtube.com/c/ProgrammingKnowledge/videos')
            if c.videos:
                found = this
                break
        except:
            continue
    print("---")
    if not found:
        print(
            f"https://www.youtube.com/{channel_id} (@{channel_name}) did not work")
        return None
    video_count = len(c.videos)
    print(
        f"Now working on {video_count} videos from {channel_name} at {found}", flush=True)
    # TODO: methods to skip a channel
    # like:
    # check if newest video is already in files (what if you lost some other or an older got unblocked or sth)
    # check if video_count == filecount in both audio and info folder (what if an older got delted, but a new one came out)
    info_count = len(os.listdir(infos_dir))
    audio_count = len(os.listdir(audio_dir))
    if video_count == info_count:
        print(
            f"We already have all {info_count} infos & {audio_count} audios.")
    # if video_count == audio_count and video_count == info_count:
        # print(f"We already have {video_count} infos & audios.")
        return None
    else:
        print(f"We currently have {info_count} infos & {audio_count} audios.")
    print_channel(c, channel_dir)
    # prios = [1935]
    for idx, video in enumerate(c.videos):
        # TODO: Quick fix to skip what has been done already.
        # If files are missing (many reasons for that), remove this
        # TODO: make a dict of: filenames that exist : (their respective URLs)
        # if idx < info_count and idx not in prios:
        #     continue
        try:
            if idx % 50 == 0:
                print(f"Videos checked:\t{idx}/{video_count}", flush=True)
            print_infos(video, infos_dir)
            # print_audio(video, audio_dir)
        except Exception as e:
            print(f"\t\tError at {idx}:")
            try:
                print("\t\t" + video.title)
            except Exception as e2:
                print(e2, flush=True)
            print(e, flush=True)
    print('Download is complete')
    return True
    # todo: Make it so same named files do not overwrite eachother



import pandas as pd
import numpy as np
import networkx as nx
import matplotlib.pyplot as plt
import math
from graphviz import Source
import os


def sample():
    # Build a dataframe with 4 connections
    df = pd.DataFrame(
        {'from': ['A', 'B', 'C', 'A'], 'to': ['D', 'A', 'E', 'C']})

    # Build your graph
    G = nx.from_pandas_edgelist(df, 'from', 'to')

    # Plot it
    nx.draw(G, with_labels=True, node_color='None', edge_color='lightblue')
    plt.show()


def graph_from_list(data):
    df = pd.DataFrame(data, columns=['from', 'to'])

    # Build your graph
    G = nx.from_pandas_edgelist(df, 'from', 'to', create_using=nx.DiGraph())
    pos = nx.spring_layout(G, k=15/math.sqrt(G.order()))
    # nx.draw_networkx(G.subgraph('root'), pos=pos, node_color='black')

    # Plot it
    # nx.draw(G, with_labels=True, node_size=1500, alpha=0.3, arrows=True)
    nx.draw(G, with_labels=True, arrows=True, node_size=1500,
            node_color='None', edge_color='lightblue')
    nx.nx_pydot.write_dot(G, 'file.dot')
    dot_plot()
    # plt.show()


def dot_plot(filename="file.dot", directory="output"):
    engines = ["dot", "fdp"]
    all_engines = ["dot", "neato", "fdp", "sfdp", "circo",
                   "twopi", "osage", "patchwork"]
    # s = Source.from_file(filename, directory=directory, encoding="utf8")
    s = Source.from_file(filename, directory=directory,
                         encoding="utf8", format='svg')
    s = s.unflatten(stagger=5)
    for engine in engines:
        try:
            s.engine = engine
            new_filename= f"{filename}_{engine}"
            s.render(filename=new_filename, directory=directory)
            if os.path.exists(os.path.join(directory, new_filename)):
                # remove the non-pdf file
                os.remove(os.path.join(directory, new_filename))
        except Exception as e:
            print(e)
    # s.view()

# dot_plot()
# sample()



# controllers.py
from __future__ import annotations
from typing import TYPE_CHECKING

import os
import numpy as np
import pandas as pd
from typing import List, Dict
import matplotlib
# If you want to debug:
matplotlib.use('Agg')
import matplotlib.pyplot as plt
from sklearn.metrics.pairwise import cosine_similarity
from scipy.spatial import distance
from wordfreq import word_frequency

if TYPE_CHECKING:
    from ..extract.nlp.util_nlp import ResearchQuestion

MAX_LABELS = 15

# todo: https://towardsdatascience.com/a-glance-at-text-similarity-52e607bfe920
# TF-IDF (term frequency-inverse document frequency)

# Word embeddings are vector representations of words that capture the semantic relationships between them.
#   Instead of comparing the frequency of individual words, you can compare the similarity of their
#   vector representations. This can provide a more nuanced understanding of the meaning of the texts.


def create_word_vectors(dictionaries: List[Dict[str, int]], sum_dict: Dict[str, int] = None):
    """
    Create episode vectors from a list of dictionaries.

    Args:
        dictionaries (List[Dict[str, int]]): A list of dictionaries, where each dictionary represents
            the word frequencies for an episode.

    Returns:
        np.ndarray: A 2D numpy array where each row represents the word vector for an episode.
    """
    if sum_dict:
        unique_words = set(word for word in sum_dict.keys())
    else:
        unique_words = set(word for d in dictionaries for word in d.keys())
    vector_size = len(unique_words)
    word_indices = {word: i for i, word in enumerate(unique_words)}
    vectors = np.zeros((len(dictionaries), vector_size))
    for i, d in enumerate(dictionaries):
        for word, count in d.items():
            index = word_indices[word]
            vectors[i][index] = count
    return vectors


def normalize(vectors, log_level=False, sqrt_level=2):
    """
    Normalize a list of vectors and reward higher word counts logarithmically.

    Parameters:
    ----------
    vectors : numpy.ndarray
        A 2D numpy array of shape (n_samples, n_features) containing the vectors to normalize.
    log : bool, optional (default=True)
        Whether to apply logarithmic scaling to the values in the input array.

    Returns:
    -------
    numpy.ndarray
        A 2D numpy array of shape (n_samples, n_features) containing the normalized vectors.

    Notes:
    -----
    This function normalizes each element in the input array by dividing it by its maximum value in its column.
    If the `log` argument is True, it applies logarithmic scaling to the values in the input array by scaling up the
    values by a large constant (9999 in this case) to avoid negative values, adding 1 to each element to avoid taking
    the logarithm of 0 or negative numbers, and applying logarithmic scaling using base 10, before dividing by 4
    to reduce the magnitude of the resulting values.

    Examples:
    --------
    >>> vectors = np.array(
63        [[0.010,  0.03,   0.004], 
285        [0.14,   0.002,  0.005], 
max        [0.232,  0.118,  0.242]])
    >>> normalize(vectors, log_level=0)
    array([[0.427,  0.262,  0.018], 
285        [0.613,  0.017,  0.023], 
           [1.0,    1.0,    1.0]])
    >>> normalize(vectors, log_level=5)
    array([[0,908,  0.855, 0.564],
           [0,947,  0.555, 0.591],
           [1.0,    1.0,    1.0]])
    """
    max_values = np.max(vectors, axis=0)
    indices_zero = max_values == 0
    max_values[indices_zero] = 1  # replacing 0s with 1s
    vectors /= max_values

    if log_level:
        # Recommended: False, not fully tested.
        # If used: ~5, between 2 and 10
        vectors *= 10 ** (log_level - 1)
        vectors += 1
        vectors = np.log10(vectors) / (log_level - 1)
    if sqrt_level:
        vectors = np.sqrt(vectors)
    return vectors

# def compute_similarity_matrix(vectors):
#     # todo: Include weighed cosine similarity
#     return cosine_similarity(vectors)


def compute_similarity_from_vectors(vector_a: List[Dict[str, int]], vector_b: List[Dict[str, int]]):
    np_vectors_a = create_word_vectors(vector_a)
    np_vectors_a = normalize(np_vectors_a)

    np_vectors_b = create_word_vectors(vector_b)
    sim_matrix = cosine_similarity(np_vectors_a, np_vectors_b)
    return sim_matrix


def compute_weighed_similarity(vector_a: List[Dict[str, int]], vector_b: List[Dict[str, int]]):
    np_vectors_a = create_word_vectors(vector_a)
    np_vectors_a = normalize(np_vectors_a)

    np_vectors_b = create_word_vectors(vector_b)
    sim_matrix = np.zeros((len(np_vectors_a), len(np_vectors_b)))
    target = np.ones(len(np_vectors_b[0]))
    for c_a, vector_a in enumerate(np_vectors_a):
        for c_b, weight in enumerate(np_vectors_b):
            if sum(vector_a * weight) == 0:
                sim_matrix[c_a][c_b] = 0
            else:
                sim_matrix[c_a][c_b] = 1 - \
                    distance.cosine(vector_a, target, weight)
    return sim_matrix


def compute_similarity_matrix(dictionaries: List[Dict[str, int]], sum_dict: Dict[str, int] = None):
    vectors = create_word_vectors(dictionaries, sum_dict)
    sim_matrix = cosine_similarity(vectors)
    return sim_matrix


def print_rq(names: list[str], sim_matrix, path, prefix="", filename="", suffix="_similarity_matrix", xlabel='Documents', ylabel='Documents', dpi=300, vmin=None, vmax=None, cmap=None, additional_info: list[ResearchQuestion] = None):
    n_documents, n_questions = sim_matrix.shape

    fig, ax = plt.subplots()
    bottom = np.zeros(n_documents)

    # every_nth_x = max(1, n_documents // MAX_LABELS)

    for i, rq in enumerate(additional_info):
        name = rq.title
        p = ax.bar(names, sim_matrix[:, 0], label=name, bottom=bottom)
        bottom += sim_matrix[:, 0]
        if n_documents < MAX_LABELS:
            ax.bar_label(p, label_type='center')

    # ax.set_xticklabels(range(1, n_documents + 1, every_nth_x), rotation=90)
    ax.set_xlabel(xlabel)

    ax.set_title('Importance to our Research question')
    ax.legend()

    # Save heatmap as PNG
    filename = prefix + filename + suffix
    png_file = os.path.join(path, filename) + '.png'
    plt.savefig(png_file, dpi=dpi)

    # Save similarity matrix as CSV
    csv_file = os.path.join(path, filename) + '.csv'
    pd.DataFrame(sim_matrix).to_csv(csv_file, index=False,
                                    header=False, sep=";", decimal=",")

    # ax.legend()

    print(f"Similarity matrix saved as {png_file} and {csv_file}.")


def print_sim_matrix(sim_matrix, path, prefix="", filename="", suffix="_similarity_matrix", xlabel='Documents', ylabel='Documents', dpi=300, vmin=None, vmax=None, cmap=None, additional_info: ResearchQuestion = None):
    # If i want to see the figure:
    # fig, ax = plt.subplots()
    fig, ax = plt.subplots()
    # plt.show()
    len_y, len_x = sim_matrix.shape
    if len_x == len_y:
        np.fill_diagonal(sim_matrix, np.nan)

    # todo: fix

    if vmin is None:
        vmin = np.nanmin(sim_matrix)
    if vmax is None:
        vmax = np.nanmax(sim_matrix) * 100
    if cmap is None:
        cmap = 'Greens'
    im = ax.imshow(sim_matrix * 100, cmap=cmap, vmin=vmin, vmax=vmax)
    # if len(sim_matrix) < 10:
    #     ax.set_xticks(range(len(sim_matrix)))
    #     ax.set_yticks(range(len(sim_matrix)))
    #     ax.set_xticklabels(range(1, len(sim_matrix)+1), rotation=90)
    #     ax.set_yticklabels(range(1, len(sim_matrix)+1), rotation=90)
    every_nth_x = max(1, len_x // MAX_LABELS)
    every_nth_y = max(1, len_y // MAX_LABELS)
    ax.set_xticks(range(0, len_x, every_nth_x))
    ax.set_yticks(range(0, len_y, every_nth_y))
    ax.set_xticklabels(range(1, len_x + 1, every_nth_x), rotation=90)
    ax.set_yticklabels(range(1, len_y + 1, every_nth_y), rotation=0)
    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)
    ax.set_title('Cosine Similarity Matrix')
    cbar = plt.colorbar(im, format='%.1f%%')
    cbar.ax.set_ylabel('Similarity (%)')
    # Save heatmap as PNG
    filename = prefix + filename + suffix
    png_file = os.path.join(path, filename) + '.png'
    plt.savefig(png_file, dpi=dpi)

    # Save similarity matrix as CSV
    csv_file = os.path.join(path, filename) + '.csv'
    pd.DataFrame(sim_matrix).to_csv(csv_file, index=False,
                                    header=False, sep=";", decimal=",")

    print(f"Similarity matrix saved as {png_file} and {csv_file}.")



from  core import helper
from review import spellcheck as s_check

import os
import re
import json
import shutil


tex_esc = {
    chr(92):   chr(92) + 'textbackslash',
    '^':   chr(92) + 'textasciicircum',
    '~':   chr(92) + 'textasciitilde',
    '&':   chr(92) + '&',
    '%':   chr(92) + '%',
    '$':   chr(92) + '$',
    '#':   chr(92) + '#',
    '_':   chr(92) + '_',
    '{':   chr(92) + '{',
    '}':   chr(92) + '}',
}


def latex_escape(title, skip=False):
    # https://tex.stackexchange.com/questions/34580/escape-character-in-latex
    for key, value in tex_esc.items():
        if key == '#' and skip:
            continue
        if key == '&' and skip:
            continue
        title = title.replace(key, value)
    return title


tex_title_esc = {
    '[':   '{[}',
    ']':   '{]}',
}


def latex_title_escape(title):
    for key, value in tex_title_esc.items():
        title = title.replace(key, value)
    return title


def add_acronyms(playlist_name, f):
    if playlist_name in s_check.acro:
        ac_dict = s_check.acro[playlist_name]
    else:
        return
    if not ac_dict:
        return

    f.write('\\begin{acronym}[MMORPG]\n')
    for ac, entries in ac_dict.items():
        line = '\t\\acro{' + ac + '}[' + ac + ']{' + entries[0] + '}'
        f.write(line + '\n')
        if len(entries) > 1:
            line = '\t\\acrodefplural{' + ac + \
                '}[' + entries[1] + ']{' + entries[2] + '}'
            f.write(line + '\n')
    f.write('\\end{acronym}\n')


def texify_acronyms(text, playlist_name):
    if playlist_name in s_check.acro:
        ac_dict = s_check.acro[playlist_name]
    else:
        return text
    if not ac_dict:
        return text

    for ac, entries in ac_dict.items():
        # Todo: find out why r'\b(?=\w)' might be used instead r'\b'
        ac_s = r'\b' + re.escape(ac) + r'\b'
        ac = r'\\ac{' + ac + '}'
        if len(entries) > 1:
            acp_s = r'\b' + re.escape(entries[1]) + r'\b'
            acp = r'\\ac{' + entries[1] + '}'
        else:
            acp_s = r'\b' + re.escape(ac) + r's\b'
            acp = r'\\ac{' + ac + 's}'
        text = re.sub(ac_s, ac, text, flags=re.IGNORECASE)
        text = re.sub(acp_s, acp, text, flags=re.IGNORECASE)
    return text


def texify(var_s, k):
    if k == 'url':
        tokens = var_s.split(' ; ')
        winner = ""
        for t in tokens:
            url_var = helper.get_brand(t)
            winner += ' \href{' + t + \
                '}{\includegraphics[height=11pt]{gfx/' + url_var + '.png}}'
        var_s = winner
    return var_s


def convert_to_tex(input_path, playlist_name, playlist_info, episodes_info, wiki_episodes_info={}):
    playlist_info, episodes_info = helper.setup_infos(
        playlist_info, episodes_info, input_path)
    language = 'de'
    if 'language' in playlist_info:
        language = playlist_info['language']
    # https://stackoverflow.com/questions/27844088/python-get-directory-two-levels-up
    two_up = os.path.abspath(os.path.join(input_path, ".."))
    tex_path_sample = two_up + '\\sample\\' + language + '\\LaTeX\\'
    tex_path = input_path + 'LaTeX\\'
    if not os.path.exists(tex_path):
        os.makedirs(tex_path)
    data_files, data_folders = helper.extract_file_folder(tex_path_sample)
    tex_filenames = []
    for filename in data_files:
        split_tup = os.path.splitext(tex_path_sample + filename)
        file_name = split_tup[0]
        file_extension = split_tup[1]
        if file_extension == '.tex':
            tex_filenames.append(filename)
    data_files = tex_filenames
    for filename in data_files:
        filename
        src = tex_path_sample + '\\' + filename
        dst = tex_path + '\\' + filename
        shutil.copyfile(src, dst)

    if 'raw' in data_folders:
        data_folders.remove('raw')
    for foldername in data_folders:
        if not os.path.exists(tex_path + '\\' + foldername):
            os.makedirs(tex_path + '\\' + foldername)

    # make changes according to podcast_info

    # Fill input
    index_lines = []
    author = ''
    if 'author_name' in playlist_info:
        author = playlist_info['author_name']
    for e_key in episodes_info:
        # min = 726
        # max = 726
        # # max = 999
        # if e_key < min:
        #     continue
        # if e_key > max:
        #     break
        episode = episodes_info[e_key]
        title = episode['title']
        clean_name = helper.get_clean_title(title, e_key)
        json_path = helper.get_edited_path(clean_name, input_path)

        # reminder: Episodes that have no mp3 (i.E. Spotify exclusives) cannot be found
        if os.path.exists(json_path):
            with open(json_path, encoding='utf-8') as json_file:
                info = json.load(json_file)

            author_here = author
            # Todo: Make this read and differentiate the episodes author
            # Currently this is not even extracted from YouTube
            if 'author' in episode:
                author_here = episode['author']
            # YouTube and Spotify
            if 'url' in episode:
                # author_here += texify(episode['url'], 'url')
                # Wenn Icons dort sind, brauchen wir das "Onkel Barlow" vielleicht nicht
                urls = episode['url']
                if e_key in wiki_episodes_info:
                    if 'url_wiki' in wiki_episodes_info[e_key]:
                        url_wiki = wiki_episodes_info[e_key]['url_wiki'].replace(
                            '%', '\\%')
                        urls = url_wiki + ' ; ' + urls
                author_here = texify(urls, 'url')
            # Wiki
                # author_here += texify(episode['url'], 'url')
            title_forged = helper.forge_title(title, e_key, playlist_name)
            title_tex = latex_escape(title_forged)
            title_tex = latex_title_escape(title_tex)
            # No BMZ in title:
            # title_tex = re.sub(r'BMZ *:*', '', title_tex, 1)

            line = '\\newchapter{' + title_tex + '}{' + author_here + '}'
            # line = '\\newchapter{' + title_tex + '}{' + author_here + '}'
            index_lines.append(line)

            index_lines.append('\\acresetall')
            # Todo: find out why double space is a no-go
            clean_name_tex = latex_escape(clean_name, True).replace("  ", " ")

            line = '\\input{' + 'input/' + clean_name_tex + '.tex' + '}'
            index_lines.append(line)

            # Todo: find out why double space is a no-go
            tex_path_episode = tex_path + 'input\\' + \
                clean_name.replace("  ", " ") + '.tex'
            with open(tex_path_episode, "w", encoding='utf8') as f:
                text = latex_escape(info['text'])
                text = texify_acronyms(text, playlist_name)
                text = text.replace(' ', '\n')
                f.write(text)

            index_lines.append('\clearpage')
            # \newchapter{Story One}{Author One}

    if len(index_lines) > 0:
        # TODO: acronyms
        with open(tex_path + 'input\\index.tex', "w", encoding='utf8') as f:
            add_acronyms(playlist_name, f)
            line = '\part{' + playlist_name + '}'
            f.write(line + '\n')
            for line in index_lines:
                f.write(line + '\n')


def main(data_pl_path, pl_n, playlist_info, episodes_info):
    print('convert_to_tex', flush=True)
    wiki_episodes_info = "I need to change how this information is carried"
    # todo: make it so that wiki_episodes_infos is no longer needed
    convert_to_tex(data_pl_path, pl_n, playlist_info,
                   episodes_info, wiki_episodes_info)


if __name__ == '__main__':
    main()



import logging
from mwclient import Site


class page_promt:
    # , **kwargs):
    def __init__(self, text, summary='', minor=False, bot=True, section=None):
        self.text = text
        self.summary = summary
        self.bot = bot
        self.section = section
        self.text = text
        # TODO: understand this
        # self.**kwargs = **kwargs


def handle_queries(queries):
    logging.basicConfig(level=logging.WARNING)

    ua = 'playlist2wiki-converter/0.0 (https://github.com/borgnetzwerk/playlist2wiki-converter)'
    site = Site('data.bnwiki.de', path='/', clients_useragent=ua)

    with open("C:\\auth_tokens\\a.txt") as f:
        pw = f.readlines()[0]
    site.login('timborg', pw)
    del pw

    for pagename, tasks in queries.items():
        page = site.pages[pagename]
        for order, promt in tasks.items():
            if order == 'edit':
                if page.text() == promt.text:
                    continue
                """If the page didn’t exist, this operation will create it."""
                page.edit(promt.text, summary=promt.summary,
                          bot=promt.bot, section=promt.section)
            else:
                print('unknown order: ' + str(order))
            # if page.exists:
            # page.exists
            # text = page.text()


def main(queries):
    handle_queries(queries)

    # csv2table(input_path, lookupCSV(input_path))


if __name__ == '__main__':
    main()



from ...core import helper
from wiki import main as wiki_helpers
import wiki_bot

import json
import csv
import sys
import os
from os import listdir
from os.path import isfile, join
import urllib.parse


# from ...wiki_bot.src import main as wiki_bot
# https://iq-inc.com/importerror-attempted-relative-import/
# https://techwithtech.com/importerror-attempted-relative-import-with-no-known-parent-package/
# https://stackoverflow.com/questions/7505988/importing-from-a-relative-path-in-python

# first change the cwd to the script path
# scriptPath = os.path.realpath(os.path.dirname(sys.argv[0]))
# os.chdir(scriptPath)

# #append the relative location you want to import from
# sys.path.append("../")
# import wiki_bot.src.main as wiki_bot
# sys.path.remove('/wiki_bot/src')
# import your module stored in '../common'


csv_delimiter = ';'

individual_wiki_pages_per_episode = False

"""
old functionality can be restored with:
write_individual_wiki_page

"""


def wiki_forge_title(title, playlist_name, id):
    if id is not None:
        title = helper.forge_title(title, id, playlist_name)
    title = title.replace('[', '(')
    title = title.replace(']', ')')
    title = title.replace('#', '')
    return title


def wikify_dict(dict, playlist_name, id=None, author_name='TBD'):
    temp = {}
    for k in dict:
        dict[k] = helper.remove_quot(dict[k])
        if k == 'id':
            temp[k] = dict[k]
        elif k == 'date':
            dict[k] = helper.time_converter(dict[k])
            temp[k] = dict[k]
        elif k == 'runtime':
            dict[k] = helper.time_converter(dict[k])
            temp[k] = dict[k]
        elif k == 'name':
            temp[k] = dict[k]
        elif k == 'title':
            temp['title_raw'] = dict[k]
            title = wiki_forge_title(dict[k], playlist_name, id)
            title_full = '/'.join([author_name, playlist_name, title])
            link_to_title = '[[' + title_full + '|' + title + ']]'
            temp['title'] = title
            temp['title_full'] = title_full
            temp['link_to_title'] = link_to_title
        elif k == 'description':
            temp[k] = dict[k]
        elif k == 'author':
            temp2 = helper.cut_out(dict[k], helper.dict_author)
            temp['author_type'] = helper.remove_quot(temp2['author_type'])
            temp['author_name'] = helper.remove_quot(temp2['author_name'])
        elif k == 'publisher':
            temp[k] = dict[k]
        elif k == 'language':
            temp[k] = dict[k]
        elif k == '@type':
            temp[k] = dict[k]
        elif k == 'accessMode':
            temp[k] = dict[k]
        elif k == 'url':
            tokens = dict[k].split(' ; ')
            winner = ""
            for t in tokens:
                url_var = 'Link'
                if "spotify.com" in t:
                    url_var = " Spotify"
                elif "youtube.com" in t:
                    url_var = " YouTube"
                winner += '[' + t + url_var + ']'
            temp[k] = winner
        elif k == 'image':
            temp[k] = '[' + dict[k] + ' image]'
        else:
            temp[k] = dict[k]
    return temp


def lookupCSV(input_path):
    input_path
    onlyfiles = [f for f in listdir(input_path) if isfile(join(input_path, f))]
    return [f for f in onlyfiles if ".csv" in f]


# convert from csv to wiki

def csv2wiki(playlist_info, episodes, input_path, playlist_name):
    # Könnte man auch als Übersicht über alle podcast machen, die im Data.bnwiki sind
    # Sortierbar nach Sprache etc.
    # Kategorien ...
    # mediaWiki.main(playlist_info, episodes, input_path, playlist_name)
    return

# convert from json to wiki


def json2wiki(playlist_info, episodes, input_path, filename, playlist_name):
    # Könnte man auch als Übersicht über alle podcast machen, die im Data.bnwiki sind
    # Sortierbar nach Sprache etc.
    # Kategorien ...
    # mediaWiki.main(input_path)
    return

# TODO: Rework these:


def some_old_func(playlist_info={}, episodes_info={}, input_path=os.getcwd(), playlist_name=''):
    playlist_info = wikify_dict(playlist_info, playlist_name)
    for e_key in episodes_info:
        episodes_info[e_key] = wikify_dict(episodes_info[e_key], playlist_name)
    # json2csv(playlist_info, episodes_info, input_path, playlist_name)


def add_wiki_url(wiki_episodes_info):
    for eID, episode_info in wiki_episodes_info.items():
        url_title = urllib.parse.quote(episode_info['title_full'])
        wiki_url = 'https://data.bnwiki.de/index.php?title=' + url_title
        episode_info['url_wiki'] = wiki_url
    return wiki_episodes_info


def update_wiki(playlist_info, episodes_info, input_path, playlist_name):
    wiki_playlist_info = wikify_dict(playlist_info, playlist_name)
    wiki_episodes_info = {}
    author_name = 'TBD'
    if 'author_name' in playlist_info:
        author_name = playlist_info['author_name']
    for eID, episode_info in episodes_info.items():
        wiki_episodes_info[eID] = wikify_dict(
            episode_info, playlist_name, eID, author_name)

    do_individual = True
    # do_individual = False

    queries = {}

    pagename = 'OnkelBarlow'

    text = wiki_helpers.write_episodes(
        playlist_name, wiki_episodes_info)
    tasks = {}

    order = 'edit'
    is_bot = True
    section = None
    promt = wiki_bot.page_promt(
        text, "edited from playlist2wiki-converter", is_bot, section)

    tasks.update({order: promt})

    queries.update({pagename: tasks})

    if do_individual:
        for eID, episode_info in wiki_episodes_info.items():
            # if eID > 2:
            #     break
            pagename = episode_info['title_full']

            transcript_text = ''
            clean_title = helper.get_clean_title(
                episode_info['title_raw'], eID)
            json_path = helper.get_edited_path(clean_title, input_path)

            if isfile(json_path):
                with open(json_path, encoding='utf-8') as json_file:
                    transcript = json.load(json_file)
                    # TODO: make this a better structured page:
                    transcript_text = transcript['text']
            episode_info['transcript'] = transcript_text

            episode_info['text'] = wiki_helpers.write_individual_wiki_page(
                playlist_name, eID, episode_info)
            tasks = {}

            order = 'edit'
            text = episode_info['text']
            is_bot = True
            section = None
            if 'section' in episode_info:
                section = episode_info['section']
            promt = wiki_bot.page_promt(
                text, "edited from playlist2wiki-converter", is_bot, section)

            tasks.update({order: promt})

            queries.update({pagename: tasks})
    # if len(queries) > 0:
    #     wiki_bot.main(queries)

    wiki_episodes_info = add_wiki_url(wiki_episodes_info)

    return wiki_episodes_info


def main(playlist_info={}, episodes_info={}, input_path=os.getcwd(), playlist_name=''):
    if input_path == os.getcwd():
        data_path, playlist_names = helper.get_data_folders()
        playlist_name = playlist_names[0]
        input_path = data_path + playlist_name + '\\'
        playlist_info, episodes_info = helper.setup_infos(
            playlist_info, episodes_info, input_path)
    # csv2table(input_path, lookupCSV(input_path))
    return update_wiki(playlist_info, episodes_info, input_path, playlist_name)


if __name__ == '__main__':
    main()



from core import helper

from PyPDF2 import PdfReader
import pandas as pd
import numpy as np
import math
import os
from sklearn.feature_extraction.text import CountVectorizer
from sklearn.feature_extraction.text import CountVectorizer
import codecs
import inspect
import logging

"""
    PUNISH_SEMIMATCH = 0.2
    PUNISH_NOMATCH = PUNISH_SEMIMATCH / 2
    cfg['WEAKEN_WORDCOUNT'] = 0.01
    FLAT_SCORE_BOOST = 10
}
"""


def decode_hex(hex_encoded_text):
    if not "/x" in hex_encoded_text:
        return hex_encoded_text

    # try:
    #     return bytes.fromhex(no_x_text).decode()
    # except:
    #     try:
    #         return bytes.fromhex(no_x_text).decode('latin-1')
    #     except:
    lines = hex_encoded_text.split("\n")
    line_list = []
    for text in lines:
        no_x_blocks = text.split(' ')
        blocks = []
        for block in no_x_blocks:
            if "/x" not in block or "http" in block or "www" in block:
                blocks.append(block)
                continue
            try:
                block = block.replace('/x', "")
                byte_block = bytes.fromhex(block)
            except Exception as e:
                print(f"{block}: {e}")
                try:
                    blocks.append(str(block))
                except:
                    blocks.append(" ")
                continue
            try:
                # if "\x" in block:
                try:
                    block_clean = byte_block.decode()
                    while "\\x" in block_clean:
                        block_clean = byte_block.decode('latin-1')
                        if "\\x" in block_clean:
                            print(f"couldn't resolve {block} to {block_clean}")
                # else:
                except:
                    block_clean = byte_block.decode('latin-1')
                    # blocks.append(byte_block.decode())
            except:
                block_clean = str(byte_block)[2:-1]
            blocks.append(block_clean)
        line_list.append(" ".join(blocks))
    return "\n".join(line_list)


def get_text(reader):
    fulltext = ""
    for page in reader.pages:
        text = page.extract_text()
        # some files are weirdly compiled as hex
        # That means we will decode any actual hex code in the file as well
        text = decode_hex(text)
        if fulltext:
            fulltext += "\n"
        fulltext += text
    if fulltext.count(' ') == 0:
        # todo: backup method to catch files like:
        # \files\58\Springer-TheNURBSBook.pdf
        return ""
    return fulltext


def calculate_score(entry_dict, fulltext, cfg):
    header = {'zotero_ID': entry_dict['ID']}
    scores = {}
    group_scores = {}
    appendix = {}
    total_score = 1
    wordcount = fulltext.count(" ")
    wordcount_weight = wordcount * cfg['WEAKEN_WORDCOUNT']
    for g_name, group in cfg['keywords'].items():
        scores.update({word: fulltext.count(word) for word in group})
        semimatches = []
        group_score = 0
        for word, weight in group.items():
            if not scores[word]:
                # attempt to find words otherwise
                fragments = word.split(" ")
                subscores = {fulltext.count(fragment)
                             for fragment in fragments}
                scores[word] = min(subscores) * cfg['PUNISH_SEMIMATCH']
                if scores[word] > 0:
                    semimatches.append(word)
            if scores[word] > 0:
                group_score += (scores[word]) * weight
        group_scores[g_name+"_score"] = group_score

    gsl = list(group_scores.values())
    for idx, score in enumerate(gsl):
        if score > 0:
            gsl[idx] = score/wordcount_weight
        else:
            gsl[idx] = cfg['PUNISH_NOMATCH']/wordcount_weight
    total_score = cfg['FLAT_SCORE_BOOST'] + math.log10(np.prod(gsl))
    # else:
    #     total_score = total_score/pow(wordcount, 3)
    header["total_score"] = total_score
    header.update(group_scores)
    header["wordcount"] = wordcount
    appendix["semimatches"] = ";".join(semimatches)
    entry_dict["k_score"] = header | scores | appendix
    return entry_dict


def read_pdf(filepath, entry_dict, cfg):
    # supressing errors like "FloatObject (b'99.-72') invalid; use 0.0 instead"
    # https://stackoverflow.com/questions/74866750/error-floatobject-b0-000000000000-14210855-invalid-use-0-0-instead-while-u
    logger = logging.getLogger("PyPDF2")
    logger.setLevel(logging.ERROR)

    try:
        reader = PdfReader(filepath)
    except Exception as e:
        print(e)
        return entry_dict
    fulltext = get_text(reader)
    if not fulltext:
        return entry_dict

    entry_dict = calculate_score(entry_dict, fulltext, cfg)
    return entry_dict
# read_pdf("", {})


def write_to_file(score_dict, path, cfg):
    # others
    helper.write_to_py(inspect.getsource(calculate_score),
                       os.path.join(path, "calculate_score.py"))
    # todo: make this go to yamls file
    helper.dict2json(cfg, "cfg", path)
    # CSV:
    score_dict = dict(sorted(score_dict.items(), key=lambda item: float(
        item[1]['total_score']), reverse=True))
    columns = list(list(score_dict.values())[0].keys())
    # df = pd.DataFrame([score_dict.values()], columns=columns, index=[score_dict.keys()])
    helper.write_to_csv(columns, list(score_dict.values()),
                        os.path.join(path, "k_score.csv"))
